#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2022 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
The module file for ios_lag_interfaces
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
module: ios_lag_interfaces
short_description: Resource module to configure LAG interfaces.
description:
  This module manages properties of Link Aggregation Group on Cisco IOS
  devices.
version_added: 1.0.0
author:
  - Sagar Paul (@KB-perByte)
  - Sumit Jaiswal (@justjais)
notes:
  - Tested against Cisco IOSXE Version 17.3 on CML.
  - This module works with connection C(network_cli).
    See U(https://docs.ansible.com/ansible/latest/network/user_guide/platform_ios.html)
options:
  config:
    description: A list of link aggregation group configurations.
    type: list
    elements: dict
    suboptions:
      name:
        description:
          - ID of Ethernet Channel of interfaces.
          - Refer to vendor documentation for valid port values.
        type: str
        required: true
      members:
        description:
          - Interface options for the link aggregation group.
        type: list
        elements: dict
        suboptions:
          member:
            description:
              - Interface member of the link aggregation group.
            type: str
          mode:
            description:
              - Etherchannel Mode of the interface for link aggregation.
              - On mode has to be quoted as 'on' or else pyyaml will convert
                to True before it gets to Ansible.
            type: str
            choices:
              - auto
              - "on"
              - desirable
              - active
              - passive
          link:
            description:
              - Assign a link identifier used for load-balancing.
              - Refer to vendor documentation for valid values.
              - NOTE, parameter only supported on Cisco IOS XE platform.
            type: int
  running_config:
    description:
      - This option is used only with state I(parsed).
      - The value of this option should be the output received from the IOS device
        by executing the command B(show running-config | section ^interface).
      - The state I(parsed) reads the configuration from C(running_config) option and
        transforms it into Ansible structured data as per the resource module's argspec
        and the value is then returned in the I(parsed) key within the result.
    type: str
  state:
    description:
      - The state the configuration should be left in
      - The states I(rendered), I(gathered) and I(parsed) does not perform any change
        on the device.
      - The state I(rendered) will transform the configuration in C(config) option to
        platform specific CLI commands which will be returned in the I(rendered) key
        within the result. For state I(rendered) active connection to remote host is
        not required.
      - The state I(gathered) will fetch the running configuration from device and transform
        it into structured data in the format as per the resource module argspec and
        the value is returned in the I(gathered) key within the result.
      - The state I(parsed) reads the configuration from C(running_config) option and
        transforms it into JSON format as per the resource module parameters and the
        value is returned in the I(parsed) key within the result. The value of C(running_config)
        option should be the same format as the output of command I(show running-config
        | include ip route|ipv6 route) executed on device. For state I(parsed) active
        connection to remote host is not required.
    type: str
    choices:
      - merged
      - replaced
      - overridden
      - deleted
      - rendered
      - parsed
      - gathered
    default: merged
"""

EXAMPLES = """
# Using merged
#
# Before state:
# -------------
#
# vios#show running-config | section ^interface
# interface Port-channel10
# interface GigabitEthernet0/1
#  shutdown
# interface GigabitEthernet0/2
#  shutdown
# interface GigabitEthernet0/3
#  shutdown
# interface GigabitEthernet0/4
#  shutdown

- name: Merge provided configuration with device configuration
  cisco.ios.ios_lag_interfaces:
    config:
      - name: Port-channel10
        members:
          - member: GigabitEthernet0/1
            mode: auto
          - member: GigabitEthernet0/2
            mode: auto
      - name: Port-channel20
        members:
          - member: GigabitEthernet0/3
            mode: "on"
      - name: Port-channel30
        members:
          - member: GigabitEthernet0/4
            mode: active
    state: merged

# Task Output:
# ---------------

# commands:
# - interface GigabitEthernet0/1
# - channel-group 10 mode auto
# - interface GigabitEthernet0/2
# - channel-group 10 mode auto
# - interface GigabitEthernet0/3
# - channel-group 20 mode on
# - interface GigabitEthernet0/4
# - channel-group 30 mode active

# After state:
# ------------
#
# vios#show running-config | section ^interface
# interface Port-channel10
# interface Port-channel20
# interface Port-channel30
# interface GigabitEthernet0/1
#  shutdown
#  channel-group 10 mode auto
# interface GigabitEthernet0/2
#  shutdown
#  channel-group 10 mode auto
# interface GigabitEthernet0/3
#  shutdown
#  channel-group 20 mode on
# interface GigabitEthernet0/4
#  shutdown
#  channel-group 30 mode active

# Using overridden
#
# Before state:
# -------------
#
# vios#show running-config | section ^interface
# interface Port-channel10
# interface Port-channel20
# interface Port-channel30
# interface GigabitEthernet0/1
#  shutdown
#  channel-group 10 mode auto
# interface GigabitEthernet0/2
#  shutdown
#  channel-group 10 mode auto
# interface GigabitEthernet0/3
#  shutdown
#  channel-group 20 mode on
# interface GigabitEthernet0/4
#  shutdown
#  channel-group 30 mode active

- name: Override device configuration of all interfaces with provided configuration
  cisco.ios.ios_lag_interfaces:
    config:
      - name: Port-channel20
        members:
          - member: GigabitEthernet0/2
            mode: auto
          - member: GigabitEthernet0/3
            mode: auto
    state: overridden

# Task Output:
# ---------------

# commands:
# - interface GigabitEthernet0/1
# - no channel-group 10 mode auto
# - interface GigabitEthernet0/2
# - no channel-group 10 mode auto
# - interface GigabitEthernet0/4
# - no channel-group 30 mode active
# - interface GigabitEthernet0/2
# - channel-group 20 mode auto
# - interface GigabitEthernet0/3
# - channel-group 20 mode auto

# After state:
# ------------
#
# vios#show running-config | section ^interface
# interface Port-channel10
# interface Port-channel20
# interface Port-channel30
# interface GigabitEthernet0/1
#  shutdown
# interface GigabitEthernet0/2
#  shutdown
#  channel-group 20 mode auto
# interface GigabitEthernet0/3
#  shutdown
#  channel-group 20 mode auto
# interface GigabitEthernet0/4
#  shutdown

# Using replaced
#
# Before state:
# -------------
#
# vios#show running-config | section ^interface
# interface Port-channel10
# interface Port-channel20
# interface Port-channel30
# interface GigabitEthernet0/1
#  shutdown
#  channel-group 10 mode auto
# interface GigabitEthernet0/2
#  shutdown
#  channel-group 10 mode auto
# interface GigabitEthernet0/3
#  shutdown
#  channel-group 20 mode on
# interface GigabitEthernet0/4
#  shutdown
#  channel-group 30 mode active

- name: Replaces device configuration of listed interfaces with provided configuration
  cisco.ios.ios_lag_interfaces:
    config:
      - name: Port-channel30
        members:
          - member: GigabitEthernet0/3
            mode: auto
    state: replaced

# Task Output:
# ---------------

# commands:
# - interface GigabitEthernet0/3
# - channel-group 30 mode auto
# - interface GigabitEthernet0/4
# - no channel-group 30 mode active

# After state:
# ------------
#
# vios#show running-config | section ^interface
# interface Port-channel10
# interface Port-channel20
# interface Port-channel30
# interface GigabitEthernet0/1
#  shutdown
#  channel-group 10 mode auto
# interface GigabitEthernet0/2
#  shutdown
#  channel-group 10 mode auto
# interface GigabitEthernet0/3
#  shutdown
#  channel-group 30 mode auto
# interface GigabitEthernet0/4
#  shutdown

# Using Deleted
#
# Before state:
# -------------
#
# vios#show running-config | section ^interface
# interface Port-channel10
# interface Port-channel20
# interface Port-channel30
# interface GigabitEthernet0/1
#  shutdown
#  channel-group 10 mode auto
# interface GigabitEthernet0/2
#  shutdown
#  channel-group 10 mode auto
# interface GigabitEthernet0/3
#  shutdown
#  channel-group 20 mode on
# interface GigabitEthernet0/4
#  shutdown
#  channel-group 30 mode active

- name: "Delete LAG attributes of given interfaces (Note: This won't delete the interface itself)"
  cisco.ios.ios_lag_interfaces:
    config:
      - name: Port-channel10
      - name: Port-channel20
    state: deleted

# Task Output:
# ---------------

# commands:
# - interface GigabitEthernet0/1
# - no channel-group 10 mode auto
# - interface GigabitEthernet0/2
# - no channel-group 10 mode auto
# - interface GigabitEthernet0/3
# - no channel-group 20 mode on

# After state:
# -------------
#
# vios#show running-config | section ^interface
# interface Port-channel10
# interface Port-channel20
# interface Port-channel30
# interface GigabitEthernet0/1
#  shutdown
# interface GigabitEthernet0/2
#  shutdown
# interface GigabitEthernet0/3
#  shutdown
# interface GigabitEthernet0/4
#  shutdown
#  channel-group 30 mode active

# Using Deleted without any config passed
# "(NOTE: This will delete all of configured LLDP module attributes)"

#
# Before state:
# -------------
#
# vios#show running-config | section ^interface
# interface Port-channel10
# interface Port-channel20
# interface Port-channel30
# interface GigabitEthernet0/1
#  shutdown
#  channel-group 10 mode auto
# interface GigabitEthernet0/2
#  shutdown
#  channel-group 10 mode auto
# interface GigabitEthernet0/3
#  shutdown
#  channel-group 20 mode on
# interface GigabitEthernet0/4
#  shutdown
#  channel-group 30 mode active

- name: "Delete all configured LAG attributes for interfaces (Note: This won't delete the interface itself)"
  cisco.ios.ios_lag_interfaces:
    state: deleted

# Task Output:
# ---------------

# commands:
# - interface GigabitEthernet0/1
# - no channel-group 10 mode auto
# - interface GigabitEthernet0/2
# - no channel-group 10 mode auto
# - interface GigabitEthernet0/3
# - no channel-group 20 mode on
# - interface GigabitEthernet0/4
# - no channel-group 30 mode active

# After state:
# -------------
#
# vios#show running-config | section ^interface
# interface Port-channel10
# interface Port-channel20
# interface Port-channel30
# interface GigabitEthernet0/1
#  shutdown
# interface GigabitEthernet0/2
#  shutdown
# interface GigabitEthernet0/3
#  shutdown
# interface GigabitEthernet0/4
#  shutdown

# Using Gathered

# Before state:
# -------------
#
# vios#show running-config | section ^interface
# interface Port-channel10
# interface Port-channel20
# interface Port-channel30
# interface GigabitEthernet0/1
#   shutdown
#   channel-group 10 mode auto
# interface GigabitEthernet0/2
#   shutdown
#   channel-group 10 mode auto
# interface GigabitEthernet0/3
#   shutdown
#   channel-group 20 mode on
# interface GigabitEthernet0/4
#   shutdown
#   channel-group 30 mode active

- name: Gather listed LAG interfaces with provided configurations
  cisco.ios.ios_lag_interfaces:
    config:
    state: gathered

# Module Execution Result:
# ------------------------
#
# "gathered": [
# {
#     "members": [
#         {
#             "member": "GigabitEthernet0/1",
#             "mode": "auto"
#         },
#         {
#             "member": "GigabitEthernet0/2",
#             "mode": "auto"
#         }
#     ],
#     "name": "Port-channel10"
# },
# {
#     "members": [
#         {
#             "member": "GigabitEthernet0/3",
#             "mode": "on"
#         }
#     ],
#     "name": "Port-channel20"
# },
# {
#     "members": [
#         {
#             "member": "GigabitEthernet0/4",
#             "mode": "active"
#         }
#     ],
#     "name": "Port-channel30"
# }
# ]

# After state:
# ------------
#
# vios#sh running-config | section ^interface
# interface Port-channel10
# interface Port-channel20
# interface Port-channel30
# interface GigabitEthernet0/1
#   shutdown
#   channel-group 10 mode auto
# interface GigabitEthernet0/2
#   shutdown
#   channel-group 10 mode auto
# interface GigabitEthernet0/3
#   shutdown
#   channel-group 20 mode on
# interface GigabitEthernet0/4
#   shutdown
#   channel-group 30 mode active

# Using Rendered

- name: Render the commands for provided  configuration
  cisco.ios.ios_lag_interfaces:
    config:
      - name: Port-channel11
        members:
          - member: GigabitEthernet0/1
            mode: active
      - name: Port-channel22
        members:
          - member: GigabitEthernet0/2
            mode: passive
    state: rendered

# Module Execution Result:
# ------------------------
#
# "rendered": [
#         "interface GigabitEthernet0/1",
#         "channel-group 11 mode active",
#         "interface GigabitEthernet0/2",
#         "channel-group 22 mode passive",
#     ]

# Using Parsed

#  File: parsed.cfg
# ----------------
#
# interface GigabitEthernet0/1
# channel-group 11 mode active
# interface GigabitEthernet0/2
# channel-group 22 mode passive

- name: Parse the commands for provided configuration
  cisco.ios.ios_lag_interfaces:
    running_config: "{{ lookup('file', 'parsed.cfg') }}"
    state: parsed

# Module Execution Result:
# ------------------------
#
# "parsed": [
#     {
#         "members": [
#             {
#                 "member": "GigabitEthernet0/1",
#                 "mode": "active"
#             }
#         ],
#         "name": "Port-channel11"
#     },
#     {
#         "members": [
#             {
#                 "member": "GigabitEthernet0/2",
#                 "mode": "passive"
#             }
#         ],
#         "name": "Port-channel22"
#     }
# ]
"""

RETURN = """
before:
  description: The configuration prior to the module execution.
  returned: when I(state) is C(merged), C(replaced), C(overridden), C(deleted) or C(purged)
  type: dict
  sample: >
    This output will always be in the same format as the
    module argspec.
after:
  description: The resulting configuration after module execution.
  returned: when changed
  type: dict
  sample: >
    This output will always be in the same format as the
    module argspec.
commands:
  description: The set of commands pushed to the remote device.
  returned: when I(state) is C(merged), C(replaced), C(overridden), C(deleted) or C(purged)
  type: list
  sample:
    - interface GigabitEthernet0/1
    - channel-group 10 mode auto
    - channel-group 10 mode active link 20
rendered:
  description: The provided configuration in the task rendered in device-native format (offline).
  returned: when I(state) is C(rendered)
  type: list
  sample:
    - interface GigabitEthernet0/2
    - channel-group 20 mode auto
    - channel-group 20 mode active link 60
gathered:
  description: Facts about the network resource gathered from the remote device as structured data.
  returned: when I(state) is C(gathered)
  type: list
  sample: >
    This output will always be in the same format as the
    module argspec.
parsed:
  description: The device native config provided in I(running_config) option parsed into structured data as per module argspec.
  returned: when I(state) is C(parsed)
  type: list
  sample: >
    This output will always be in the same format as the
    module argspec.
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.cisco.ios.plugins.module_utils.network.ios.argspec.lag_interfaces.lag_interfaces import (
    Lag_InterfacesArgs,
)
from ansible_collections.cisco.ios.plugins.module_utils.network.ios.config.lag_interfaces.lag_interfaces import (
    Lag_interfaces,
)


def main():
    """
    Main entry point for module execution

    :returns: the result form module invocation
    """
    module = AnsibleModule(
        argument_spec=Lag_InterfacesArgs.argument_spec,
        mutually_exclusive=[["config", "running_config"]],
        required_if=[
            ["state", "merged", ["config"]],
            ["state", "replaced", ["config"]],
            ["state", "overridden", ["config"]],
            ["state", "rendered", ["config"]],
            ["state", "parsed", ["running_config"]],
        ],
        supports_check_mode=True,
    )

    result = Lag_interfaces(module).execute_module()
    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2024, Samita Bhattacharjee (@samiib) <samita@cisco.com>

# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_template_anp_epg_annotation
short_description: Manage EPG Annotations on Cisco Nexus Dashboard Orchestrator (NDO).
description:
- Manage Endpoint Group (EPG) Annotations on Cisco Nexus Dashboard Orchestrator (NDO).
- This module is only supported on ND v3.0 (NDO v4.2) and later.
author:
- Samita Bhattacharjee (@samiib)
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: true
  template:
    description:
    - The name of the template.
    type: str
    required: true
  anp:
    description:
    - The name of the Application Network Profile (ANP).
    type: str
    required: true
  epg:
    description:
    - The name of the EPG to manage.
    type: str
    required: true
  annotation_key:
    description:
    - The key of the Annotation object.
    type: str
    aliases: [ key ]
  annotation_value:
    description:
    - The value of the Annotation object.
    type: str
    aliases: [ value ]
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment: cisco.mso.modules
notes:
- The O(schema), O(template), O(anp) and O(epg) must exist before using this module in your playbook.
  Use M(cisco.mso.mso_schema_template) to create the schema and template.
  Use M(cisco.mso.mso_schema_template_anp) to create the ANP.
  Use M(cisco.mso.mso_schema_template_anp_epg) to create the EPG.
seealso:
- module: cisco.mso.mso_schema_template
- module: cisco.mso.mso_schema_template_anp
- module: cisco.mso.mso_schema_template_anp_epg
"""

EXAMPLES = r"""
- name: Add an annotation with key and value
  cisco.mso.mso_schema_template_anp_epg_annotation:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    anp: ANP 1
    epg: EPG 1
    annotation_key: annotation_key_1
    annotation_value: annotation_value_1
    state: present

- name: Update an annotation value with key
  cisco.mso.mso_schema_template_anp_epg_annotation:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    anp: ANP 1
    epg: EPG 1
    annotation_key: annotation_key_1
    annotation_value: annotation_value_1_updated
    state: present

- name: Query a specific annotation with key
  cisco.mso.mso_schema_template_anp_epg_annotation:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    anp: ANP 1
    epg: EPG 1
    annotation_key: annotation_key_1
    state: query
  register: query_one

- name: Query all annotations
  cisco.mso.mso_schema_template_anp_epg_annotation:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    anp: ANP 1
    epg: EPG 1
    state: query
  register: query_all

- name: Delete an annotation
  cisco.mso.mso_schema_template_anp_epg_annotation:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    anp: ANP 1
    epg: EPG 1
    annotation_key: annotation_key_1
    state: absent
"""

RETURN = r"""
"""

import copy
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec
from ansible_collections.cisco.mso.plugins.module_utils.schema import MSOSchema


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str", required=True),
        template=dict(type="str", required=True),
        anp=dict(type="str", required=True),
        epg=dict(type="str", required=True),
        annotation_key=dict(type="str", aliases=["key"], no_log=False),
        annotation_value=dict(type="str", aliases=["value"]),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["annotation_key"]],
            ["state", "present", ["annotation_key", "annotation_value"]],
        ],
    )

    schema = module.params.get("schema")
    template = module.params.get("template")
    anp = module.params.get("anp")
    epg = module.params.get("epg")
    annotation_key = module.params.get("annotation_key")
    annotation_value = module.params.get("annotation_value")
    state = module.params.get("state")

    mso = MSOModule(module)

    mso_schema = MSOSchema(mso, schema, template)
    mso_schema.set_template(template)
    mso_schema.set_template_anp(anp)
    mso_schema.set_template_anp_epg(epg)

    if annotation_key:
        mso_schema.set_template_anp_epg_annotation(annotation_key, False)
        annotation = mso_schema.schema_objects.get("template_anp_epg_annotation")
        if annotation is not None:
            mso.existing = mso.previous = copy.deepcopy(annotation.details)  # Query a specific Annotation
    else:
        epg_object = mso_schema.schema_objects["template_anp_epg"]
        mso.existing = epg_object.details.get("tagAnnotations", [])  # Query all

    path = "/templates/{0}/anps/{1}/epgs/{2}/tagAnnotations".format(template, anp, epg)

    ops = []
    if state == "present":
        mso_values = {"tagKey": annotation_key, "tagValue": annotation_value}
        mso.sanitize(mso_values)
        if mso.existing:
            if annotation_value is not None and mso.existing.get("tagValue") != annotation_value:
                ops.append({"op": "replace", "path": "{0}/{1}".format(path, annotation.index), "value": mso_values})

        else:
            ops.append({"op": "add", "path": "{0}/-".format(path), "value": mso_values})
    elif state == "absent":
        if mso.existing:
            ops.append({"op": "remove", "path": "{0}/{1}".format(path, annotation.index)})

    if not module.check_mode and ops:
        mso.request(mso_schema.path, method="PATCH", data=ops)
        mso.existing = mso.proposed
    elif module.check_mode and state != "query":  # When the state is present/absent with check mode
        mso.existing = mso.proposed if state == "present" else {}

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, Kevin Breit (@kbreit) <kevin.breit@kevinbreit.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    "status": ['deprecated'],
    'supported_by': 'community'
}

DOCUMENTATION = r'''
---
module: meraki_mx_uplink_bandwidth
short_description: Manage uplinks on Meraki MX appliances
version_added: "1.1.0"
description:
- Configure and query information about uplinks on Meraki MX appliances.
notes:
- Some of the options are likely only used for developers within Meraki.
- Module was formerly named M(cisco.meraki.meraki_mx_uplink).
deprecated:
  removed_in: '3.0.0'
  why: Updated modules released with increased functionality
  alternative: cisco.meraki.networks_appliance_traffic_shaping_uplink_bandwidth
options:
    state:
      description:
      - Specifies whether object should be queried, created/modified, or removed.
      choices: [absent, present, query]
      default: query
      type: str
    org_name:
        description:
        - Name of organization associated to a network.
        type: str
    org_id:
        description:
        - ID of organization associated to a network.
        type: str
    net_name:
      description:
      - Name of network which VLAN is in or should be in.
      aliases: [network]
      type: str
    net_id:
      description:
      - ID of network which VLAN is in or should be in.
      type: str
    wan1:
      description:
      - Configuration of WAN1 uplink
      type: dict
      suboptions:
        bandwidth_limits:
          description:
          - Structure for configuring bandwidth limits
          type: dict
          suboptions:
            limit_up:
              description:
              - Maximum upload speed for interface
              type: int
            limit_down:
              description:
              - Maximum download speed for interface
              type: int
    wan2:
      description:
      - Configuration of WAN2 uplink
      type: dict
      suboptions:
        bandwidth_limits:
          description:
          - Structure for configuring bandwidth limits
          type: dict
          suboptions:
            limit_up:
              description:
              - Maximum upload speed for interface
              type: int
            limit_down:
              description:
              - Maximum download speed for interface
              type: int
    cellular:
      description:
      - Configuration of cellular uplink
      type: dict
      suboptions:
        bandwidth_limits:
          description:
          - Structure for configuring bandwidth limits
          type: dict
          suboptions:
            limit_up:
              description:
              - Maximum upload speed for interface
              type: int
            limit_down:
              description:
              - Maximum download speed for interface
              type: int
author:
- Kevin Breit (@kbreit)
extends_documentation_fragment: cisco.meraki.meraki
'''

EXAMPLES = r'''
- name: Set MX uplink settings
  meraki_mx_uplink_bandwidth:
    auth_key: '{{ auth_key }}'
    state: present
    org_name: '{{ test_org_name }}'
    net_name: '{{ test_net_name }} - Uplink'
    wan1:
      bandwidth_limits:
        limit_down: 1000000
        limit_up: 1000
    cellular:
      bandwidth_limits:
        limit_down: 0
        limit_up: 0
  delegate_to: localhost

- name: Query MX uplink settings
  meraki_mx_uplink_bandwidth:
    auth_key: '{{ auth_key }}'
    state: query
    org_name: '{{ test_org_name }}'
    net_name: '{{ test_net_name }} - Uplink'
  delegate_to: localhost
'''

RETURN = r'''

data:
  description: Information about the organization which was created or modified
  returned: success
  type: complex
  contains:
    wan1:
      description: WAN1 interface
      returned: success
      type: complex
      contains:
        bandwidth_limits:
          description: Structure for uplink bandwidth limits
          returned: success
          type: complex
          contains:
            limit_up:
              description: Upload bandwidth limit
              returned: success
              type: int
            limit_down:
              description: Download bandwidth limit
              returned: success
              type: int
    wan2:
      description: WAN2 interface
      returned: success
      type: complex
      contains:
        bandwidth_limits:
          description: Structure for uplink bandwidth limits
          returned: success
          type: complex
          contains:
            limit_up:
              description: Upload bandwidth limit
              returned: success
              type: int
            limit_down:
              description: Download bandwidth limit
              returned: success
              type: int
    cellular:
      description: cellular interface
      returned: success
      type: complex
      contains:
        bandwidth_limits:
          description: Structure for uplink bandwidth limits
          returned: success
          type: complex
          contains:
            limit_up:
              description: Upload bandwidth limit
              returned: success
              type: int
            limit_down:
              description: Download bandwidth limit
              returned: success
              type: int
'''

from ansible.module_utils.basic import AnsibleModule, json
from ansible.module_utils.common.dict_transformations import recursive_diff
from ansible_collections.cisco.meraki.plugins.module_utils.network.meraki.meraki import MerakiModule, meraki_argument_spec

INT_NAMES = ('wan1', 'wan2', 'cellular')


def clean_custom_format(data):
    for interface in data:
        if data[interface]['bandwidth_limits']['limit_up'] is None:
            data[interface]['bandwidth_limits']['limit_up'] = 0
        if data[interface]['bandwidth_limits']['limit_down'] is None:
            data[interface]['bandwidth_limits']['limit_down'] = 0
    return data


def meraki_struct_to_custom_format(data):
    new_struct = {}
    for interface in INT_NAMES:
        if interface in data['bandwidthLimits']:
            new_struct[interface] = {'bandwidth_limits': {'limit_up': data['bandwidthLimits'][interface]['limitUp'],
                                                          'limit_down': data['bandwidthLimits'][interface]['limitDown'],
                                                          }
                                     }
    # return snake_dict_to_camel_dict(new_struct)
    return new_struct


def main():
    # define the available arguments/parameters that a user can pass to
    # the module

    bandwidth_arg_spec = dict(limit_up=dict(type='int'),
                              limit_down=dict(type='int'),
                              )

    interface_arg_spec = dict(bandwidth_limits=dict(type='dict', default=None, options=bandwidth_arg_spec),
                              )

    argument_spec = meraki_argument_spec()
    argument_spec.update(state=dict(type='str', choices=['absent', 'present', 'query'], default='query'),
                         net_name=dict(type='str', aliases=['network']),
                         net_id=dict(type='str'),
                         wan1=dict(type='dict', default=None, options=interface_arg_spec),
                         wan2=dict(type='dict', default=None, options=interface_arg_spec),
                         cellular=dict(type='dict', default=None, options=interface_arg_spec),
                         )

    # the AnsibleModule object will be our abstraction working with Ansible
    # this includes instantiation, a couple of common attr would be the
    # args/params passed to the execution, as well as if the module
    # supports check mode
    module = AnsibleModule(argument_spec=argument_spec,
                           supports_check_mode=True,
                           )
    meraki = MerakiModule(module, function='mx_uplink')

    meraki.params['follow_redirects'] = 'all'

    query_urls = {'mx_uplink': '/networks/{net_id}/appliance/trafficShaping/uplinkBandwidth'}
    update_bw_url = {'mx_uplink': '/networks/{net_id}/appliance/trafficShaping/uplinkBandwidth'}

    meraki.url_catalog['get_all'].update(query_urls)
    meraki.url_catalog['update_bw'] = update_bw_url

    payload = dict()

    org_id = meraki.params['org_id']
    if org_id is None:
        org_id = meraki.get_org_id(meraki.params['org_name'])
    net_id = meraki.params['net_id']
    if net_id is None:
        nets = meraki.get_nets(org_id=org_id)
        net_id = meraki.get_net_id(net_name=meraki.params['net_name'], data=nets)

    if meraki.params['state'] == 'query':
        path = meraki.construct_path('get_all', net_id=net_id)
        response = meraki.request(path, method='GET')
        data = clean_custom_format(meraki_struct_to_custom_format(response))
        meraki.result['data'] = data
    elif meraki.params['state'] == 'present':
        path = meraki.construct_path('get_all', net_id=net_id)
        original = meraki.request(path, method='GET')
        payload = {'bandwidthLimits': {}}
        for interface in INT_NAMES:
            if meraki.params[interface] is not None:
                if meraki.params[interface]['bandwidth_limits'] is not None:
                    payload['bandwidthLimits'][interface] = None
                    payload['bandwidthLimits'][interface] = {'limitUp': meraki.params[interface]['bandwidth_limits']['limit_up'],
                                                             'limitDown': meraki.params[interface]['bandwidth_limits']['limit_down'],
                                                             }
                    if payload['bandwidthLimits'][interface]['limitUp'] == 0:
                        payload['bandwidthLimits'][interface]['limitUp'] = None
                    if payload['bandwidthLimits'][interface]['limitDown'] == 0:
                        payload['bandwidthLimits'][interface]['limitDown'] = None
        if meraki.is_update_required(original, payload):
            if meraki.module.check_mode is True:
                diff = recursive_diff(clean_custom_format(meraki_struct_to_custom_format(original)),
                                      clean_custom_format(meraki_struct_to_custom_format(payload)))
                original.update(payload)
                meraki.result['data'] = clean_custom_format(meraki_struct_to_custom_format(original))
                meraki.result['diff'] = {'before': diff[0],
                                         'after': diff[1],
                                         }
                meraki.result['changed'] = True
                meraki.exit_json(**meraki.result)
            path = meraki.construct_path('update_bw', net_id=net_id)
            response = meraki.request(path, method='PUT', payload=json.dumps(payload))
            if meraki.status == 200:
                formatted_original = clean_custom_format(meraki_struct_to_custom_format(original))
                formatted_response = clean_custom_format(meraki_struct_to_custom_format(response))
                diff = recursive_diff(formatted_original, formatted_response)
                meraki.result['diff'] = {'before': diff[0],
                                         'after': diff[1],
                                         }
                meraki.result['data'] = formatted_response
                meraki.result['changed'] = True
        else:
            meraki.result['data'] = clean_custom_format(meraki_struct_to_custom_format(original))

    # in the event of a successful module execution, you will want to
    # simple AnsibleModule.exit_json(), passing the key/value results
    meraki.exit_json(**meraki.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python

# Copyright (c) 2012, Jan-Piet Mens <jpmens () gmail.com>
# Copyright (c) 2015, Ales Nosek <anosek.nosek () gmail.com>
# Copyright (c) 2017, Ansible Project
# Copyright (c) 2023, Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: ini_file
short_description: Tweak settings in INI files
extends_documentation_fragment:
  - ansible.builtin.files
  - community.general._attributes
description:
  - Manage (add, remove, change) individual settings in an INI-style file without having to manage the file as a whole with,
    say, M(ansible.builtin.template) or M(ansible.builtin.assemble).
  - Adds missing sections if they do not exist.
  - This module adds missing ending newlines to files to keep in line with the POSIX standard, even when no other modifications
    need to be applied.
attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
options:
  path:
    description:
      - Path to the INI-style file; this file is created if required.
    type: path
    required: true
    aliases: [dest]
  section:
    description:
      - Section name in INI file. This is added if O(state=present) automatically when a single value is being set.
      - If being omitted, the O(option) is placed before the first O(section).
      - Omitting O(section) is also required if the config format does not support sections.
    type: str
  section_has_values:
    type: list
    elements: dict
    suboptions:
      option:
        type: str
        description: Matching O(section) must contain this option.
        required: true
      value:
        type: str
        description: Matching O(section_has_values[].option) must have this specific value.
      values:
        description:
          - The string value to be associated with an O(section_has_values[].option).
          - Mutually exclusive with O(section_has_values[].value).
          - O(section_has_values[].value=v) is equivalent to O(section_has_values[].values=[v]).
        type: list
        elements: str
    description:
      - Among possibly multiple sections of the same name, select the first one that contains matching options and values.
      - With O(state=present), if a suitable section is not found, a new section is added, including the required options.
      - With O(state=absent), at most one O(section) is removed if it contains the values.
    version_added: 8.6.0
  option:
    description:
      - If set (required for changing a O(value)), this is the name of the option.
      - May be omitted if adding/removing a whole O(section).
    type: str
  value:
    description:
      - The string value to be associated with an O(option).
      - May be omitted when removing an O(option).
      - Mutually exclusive with O(values).
      - O(value=v) is equivalent to O(values=[v]).
    type: str
  values:
    description:
      - The string value to be associated with an O(option).
      - May be omitted when removing an O(option).
      - Mutually exclusive with O(value).
      - O(value=v) is equivalent to O(values=[v]).
    type: list
    elements: str
    version_added: 3.6.0
  backup:
    description:
      - Create a backup file including the timestamp information so you can get the original file back if you somehow clobbered
        it incorrectly.
    type: bool
    default: false
  state:
    description:
      - If set to V(absent) and O(exclusive) set to V(true) all matching O(option) lines are removed.
      - If set to V(absent) and O(exclusive) set to V(false) the specified O(option=value) lines are removed, but the other
        O(option)s with the same name are not touched.
      - If set to V(present) and O(exclusive) set to V(false) the specified O(option=values) lines are added, but the other
        O(option)s with the same name are not touched.
      - If set to V(present) and O(exclusive) set to V(true) all given O(option=values) lines are added and the other O(option)s
        with the same name are removed.
    type: str
    choices: [absent, present]
    default: present
  exclusive:
    description:
      - If set to V(true) (default), all matching O(option) lines are removed when O(state=absent), or replaced when O(state=present).
      - If set to V(false), only the specified O(value)/O(values) are added when O(state=present), or removed when O(state=absent),
        and existing ones are not modified.
    type: bool
    default: true
    version_added: 3.6.0
  no_extra_spaces:
    description:
      - Do not insert spaces before and after '=' symbol.
    type: bool
    default: false
  ignore_spaces:
    description:
      - Do not change a line if doing so would only add or remove spaces before or after the V(=) symbol.
    type: bool
    default: false
    version_added: 7.5.0
  create:
    description:
      - If set to V(false), the module fails if the file does not already exist.
      - By default it creates the file if it is missing.
    type: bool
    default: true
  allow_no_value:
    description:
      - Allow option without value and without '=' symbol.
    type: bool
    default: false
  modify_inactive_option:
    description:
      - By default the module replaces a commented line that matches the given option.
      - Set this option to V(false) to avoid this. This is useful when you want to keep commented example C(key=value) pairs
        for documentation purposes.
    type: bool
    default: true
    version_added: 8.0.0
  follow:
    description:
      - This flag indicates that filesystem links, if they exist, should be followed.
      - O(follow=true) can modify O(path) when combined with parameters such as O(mode).
    type: bool
    default: false
    version_added: 7.1.0
notes:
  - While it is possible to add an O(option) without specifying a O(value), this makes no sense.
  - As of community.general 3.2.0, UTF-8 BOM markers are discarded when reading files.
seealso:
  - plugin: ansible.builtin.ini
    plugin_type: lookup
  - plugin: community.general.from_ini
    plugin_type: filter
  - plugin: community.general.to_ini
    plugin_type: filter
author:
  - Jan-Piet Mens (@jpmens)
  - Ales Nosek (@noseka1)
"""

EXAMPLES = r"""
- name: Ensure "fav=lemonade is in section "[drinks]" in specified file
  community.general.ini_file:
    path: /etc/conf
    section: drinks
    option: fav
    value: lemonade
    mode: '0600'
    backup: true

- name: Ensure "temperature=cold is in section "[drinks]" in specified file
  community.general.ini_file:
    path: /etc/anotherconf
    section: drinks
    option: temperature
    value: cold
    backup: true

- name: Add "beverage=lemon juice" is in section "[drinks]" in specified file
  community.general.ini_file:
    path: /etc/conf
    section: drinks
    option: beverage
    value: lemon juice
    mode: '0600'
    state: present
    exclusive: false

- name: Ensure multiple values "beverage=coke" and "beverage=pepsi" are in section "[drinks]" in specified file
  community.general.ini_file:
    path: /etc/conf
    section: drinks
    option: beverage
    values:
      - coke
      - pepsi
    mode: '0600'
    state: present

- name: Add "beverage=lemon juice" outside a section in specified file
  community.general.ini_file:
    path: /etc/conf
    option: beverage
    value: lemon juice
    state: present

- name: Remove the peer configuration for 10.128.0.11/32
  community.general.ini_file:
    path: /etc/wireguard/wg0.conf
    section: Peer
    section_has_values:
      - option: AllowedIps
        value: 10.128.0.11/32
    mode: '0600'
    state: absent

- name: Add "beverage=lemon juice" outside a section in specified file
  community.general.ini_file:
    path: /etc/conf
    option: beverage
    value: lemon juice
    state: present

- name: Update the public key for peer 10.128.0.12/32
  community.general.ini_file:
    path: /etc/wireguard/wg0.conf
    section: Peer
    section_has_values:
      - option: AllowedIps
        value: 10.128.0.12/32
    option: PublicKey
    value: xxxxxxxxxxxxxxxxxxxx
    mode: '0600'
    state: present

- name: Remove the peer configuration for 10.128.0.11/32
  community.general.ini_file:
    path: /etc/wireguard/wg0.conf
    section: Peer
    section_has_values:
      - option: AllowedIps
        value: 10.4.0.11/32
    mode: '0600'
    state: absent

- name: Update the public key for peer 10.128.0.12/32
  community.general.ini_file:
    path: /etc/wireguard/wg0.conf
    section: Peer
    section_has_values:
      - option: AllowedIps
        value: 10.4.0.12/32
    option: PublicKey
    value: xxxxxxxxxxxxxxxxxxxx
    mode: '0600'
    state: present
"""

import os
import re
import tempfile
import traceback

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_bytes, to_text


def match_opt(option, line):
    option = re.escape(option)
    return re.match(f"( |\t)*([#;]?)( |\t)*({option})( |\t)*(=|$)( |\t)*(.*)", line)


def match_active_opt(option, line):
    option = re.escape(option)
    return re.match(f"()()( |\t)*({option})( |\t)*(=|$)( |\t)*(.*)", line)


def update_section_line(option, changed, section_lines, index, changed_lines, ignore_spaces, newline, msg):
    option_changed = None
    if ignore_spaces:
        old_match = match_opt(option, section_lines[index])
        if not old_match.group(2):
            new_match = match_opt(option, newline)
            option_changed = old_match.group(8) != new_match.group(8)
    if option_changed is None:
        option_changed = section_lines[index] != newline
    if option_changed:
        section_lines[index] = newline
    changed = changed or option_changed
    if option_changed:
        msg = "option changed"
    changed_lines[index] = 1
    return (changed, msg)


def check_section_has_values(section_has_values, section_lines):
    if section_has_values is not None:
        for condition in section_has_values:
            for line in section_lines:
                match = match_opt(condition["option"], line)
                if match and (len(condition["values"]) == 0 or match.group(8) in condition["values"]):
                    break
            else:
                return False
    return True


def do_ini(
    module,
    filename,
    section=None,
    section_has_values=None,
    option=None,
    values=None,
    state="present",
    exclusive=True,
    backup=False,
    no_extra_spaces=False,
    ignore_spaces=False,
    create=True,
    allow_no_value=False,
    modify_inactive_option=True,
    follow=False,
):
    if section is not None:
        section = to_text(section)
    if option is not None:
        option = to_text(option)

    # deduplicate entries in values
    values_unique = []
    [values_unique.append(to_text(value)) for value in values if value not in values_unique and value is not None]
    values = values_unique

    diff = dict(
        before="",
        after="",
        before_header=f"{filename} (content)",
        after_header=f"{filename} (content)",
    )

    if follow and os.path.islink(filename):
        target_filename = os.path.realpath(filename)
    else:
        target_filename = filename

    if not os.path.exists(target_filename):
        if not create:
            module.fail_json(rc=257, msg=f"Destination {target_filename} does not exist!")
        destpath = os.path.dirname(target_filename)
        if not os.path.exists(destpath) and not module.check_mode:
            os.makedirs(destpath)
        ini_lines = []
    else:
        with open(target_filename, encoding="utf-8-sig") as ini_file:
            ini_lines = [to_text(line) for line in ini_file.readlines()]

    if module._diff:
        diff["before"] = "".join(ini_lines)

    changed = False

    # ini file could be empty
    if not ini_lines:
        ini_lines.append("\n")

    # last line of file may not contain a trailing newline
    if ini_lines[-1] == "" or ini_lines[-1][-1] != "\n":
        ini_lines[-1] += "\n"
        changed = True

    # append fake section lines to simplify the logic
    # At top:
    # Fake random section to do not match any other in the file
    # Using commit hash as fake section name
    fake_section_name = "ad01e11446efb704fcdbdb21f2c43757423d91c5"

    # Insert it at the beginning
    ini_lines.insert(0, f"[{fake_section_name}]")

    # At bottom:
    ini_lines.append("[")

    # If no section is defined, fake section is used
    if not section:
        section = fake_section_name

    within_section = not section
    section_start = section_end = 0
    msg = "OK"
    sep = "=" if no_extra_spaces else " = "

    option_no_value_present = False

    non_blank_non_comment_pattern = re.compile(to_text(r"^[ \t]*([#;].*)?$"))

    before = after = []
    section_lines = []

    section_pattern = re.compile(to_text(rf"^\[\s*{re.escape(section.strip())}\s*]"))

    for index, line in enumerate(ini_lines):
        # end of section:
        if within_section and line.startswith("["):
            if check_section_has_values(section_has_values, ini_lines[section_start:index]):
                section_end = index
                break
            else:
                # look for another section
                within_section = False
                section_start = section_end = 0

        # find start and end of section
        if section_pattern.match(line):
            within_section = True
            section_start = index

    before = ini_lines[0:section_start]
    section_lines = ini_lines[section_start:section_end]
    after = ini_lines[section_end : len(ini_lines)]

    # Keep track of changed section_lines
    changed_lines = [0] * len(section_lines)

    # Determine whether to consider using commented out/inactive options or only active ones
    if modify_inactive_option:
        match_function = match_opt
    else:
        match_function = match_active_opt

    # handling multiple instances of option=value when state is 'present' with/without exclusive is a bit complex
    #
    # 1. edit all lines where we have a option=value pair with a matching value in values[]
    # 2. edit all the remaining lines where we have a matching option
    # 3. delete remaining lines where we have a matching option
    # 4. insert missing option line(s) at the end of the section

    if state == "present" and option:
        for index, line in enumerate(section_lines):
            if match_function(option, line):
                match = match_function(option, line)
                if values and match.group(8) in values:
                    matched_value = match.group(8)
                    if not matched_value and allow_no_value:
                        # replace existing option with no value line(s)
                        newline = f"{option}\n"
                        option_no_value_present = True
                    else:
                        # replace existing option=value line(s)
                        newline = f"{option}{sep}{matched_value}\n"
                    (changed, msg) = update_section_line(
                        option, changed, section_lines, index, changed_lines, ignore_spaces, newline, msg
                    )
                    values.remove(matched_value)
                elif not values and allow_no_value:
                    # replace existing option with no value line(s)
                    newline = f"{option}\n"
                    (changed, msg) = update_section_line(
                        option, changed, section_lines, index, changed_lines, ignore_spaces, newline, msg
                    )
                    option_no_value_present = True
                    break

    if state == "present" and exclusive and not allow_no_value:
        # override option with no value to option with value if not allow_no_value
        if len(values) > 0:
            for index, line in enumerate(section_lines):
                if not changed_lines[index] and match_function(option, line):
                    newline = f"{option}{sep}{values.pop(0)}\n"
                    (changed, msg) = update_section_line(
                        option, changed, section_lines, index, changed_lines, ignore_spaces, newline, msg
                    )
                    if len(values) == 0:
                        break
        # remove all remaining option occurrences from the rest of the section
        for index in range(len(section_lines) - 1, 0, -1):
            if not changed_lines[index] and match_function(option, section_lines[index]):
                del section_lines[index]
                del changed_lines[index]
                changed = True
                msg = "option changed"

    if state == "present":
        # insert missing option line(s) at the end of the section
        for index in range(len(section_lines), 0, -1):
            # search backwards for previous non-blank or non-comment line
            if not non_blank_non_comment_pattern.match(section_lines[index - 1]):
                if option and values:
                    # insert option line(s)
                    for element in values[::-1]:
                        # items are added backwards, so traverse the list backwards to not confuse the user
                        # otherwise some of their options might appear in reverse order for whatever fancy reason ¯\_(ツ)_/¯
                        if element is not None:
                            # insert option=value line
                            section_lines.insert(index, f"{option}{sep}{element}\n")
                            msg = "option added"
                            changed = True
                        elif element is None and allow_no_value:
                            # insert option with no value line
                            section_lines.insert(index, f"{option}\n")
                            msg = "option added"
                            changed = True
                elif option and not values and allow_no_value and not option_no_value_present:
                    # insert option with no value line(s)
                    section_lines.insert(index, f"{option}\n")
                    msg = "option added"
                    changed = True
                break

    if state == "absent":
        if option:
            if exclusive:
                # delete all option line(s) with given option and ignore value
                new_section_lines = [line for line in section_lines if not (match_active_opt(option, line))]
                if section_lines != new_section_lines:
                    changed = True
                    msg = "option changed"
                    section_lines = new_section_lines
            elif not exclusive and len(values) > 0:
                # delete specified option=value line(s)
                new_section_lines = [
                    i
                    for i in section_lines
                    if not (match_active_opt(option, i) and match_active_opt(option, i).group(8) in values)
                ]
                if section_lines != new_section_lines:
                    changed = True
                    msg = "option changed"
                    section_lines = new_section_lines
        else:
            # drop the entire section
            if section_lines:
                section_lines = []
                msg = "section removed"
                changed = True

    # reassemble the ini_lines after manipulation
    ini_lines = before + section_lines + after

    # remove the fake section line
    del ini_lines[0]
    del ini_lines[-1:]

    if not within_section and state == "present":
        ini_lines.append(f"[{section}]\n")
        msg = "section and option added"
        if section_has_values:
            for condition in section_has_values:
                if condition["option"] != option:
                    if len(condition["values"]) > 0:
                        for value in condition["values"]:
                            ini_lines.append(f"{condition['option']}{sep}{value}\n")
                    elif allow_no_value:
                        ini_lines.append(f"{condition['option']}\n")
                elif not exclusive:
                    for value in condition["values"]:
                        if value not in values:
                            values.append(value)
        if option and values:
            for value in values:
                ini_lines.append(f"{option}{sep}{value}\n")
        elif option and not values and allow_no_value:
            ini_lines.append(f"{option}\n")
        else:
            msg = "only section added"
        changed = True

    if module._diff:
        diff["after"] = "".join(ini_lines)

    backup_file = None
    if changed and not module.check_mode:
        if backup:
            backup_file = module.backup_local(target_filename)

        encoded_ini_lines = [to_bytes(line) for line in ini_lines]
        try:
            tmpfd, tmpfile = tempfile.mkstemp(dir=module.tmpdir)
            f = os.fdopen(tmpfd, "wb")
            f.writelines(encoded_ini_lines)
            f.close()
        except OSError:
            module.fail_json(msg="Unable to create temporary file %s", traceback=traceback.format_exc())

        try:
            module.atomic_move(tmpfile, os.path.abspath(target_filename))
        except OSError:
            module.ansible.fail_json(
                msg=f"Unable to move temporary file {tmpfile} to {target_filename}, IOError",
                traceback=traceback.format_exc(),
            )

    return (changed, backup_file, diff, msg)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            path=dict(type="path", required=True, aliases=["dest"]),
            section=dict(type="str"),
            section_has_values=dict(
                type="list",
                elements="dict",
                options=dict(
                    option=dict(type="str", required=True),
                    value=dict(type="str"),
                    values=dict(type="list", elements="str"),
                ),
                mutually_exclusive=[["value", "values"]],
            ),
            option=dict(type="str"),
            value=dict(type="str"),
            values=dict(type="list", elements="str"),
            backup=dict(type="bool", default=False),
            state=dict(type="str", default="present", choices=["absent", "present"]),
            exclusive=dict(type="bool", default=True),
            no_extra_spaces=dict(type="bool", default=False),
            ignore_spaces=dict(type="bool", default=False),
            allow_no_value=dict(type="bool", default=False),
            modify_inactive_option=dict(type="bool", default=True),
            create=dict(type="bool", default=True),
            follow=dict(type="bool", default=False),
        ),
        mutually_exclusive=[["value", "values"]],
        add_file_common_args=True,
        supports_check_mode=True,
    )

    path = module.params["path"]
    section = module.params["section"]
    section_has_values = module.params["section_has_values"]
    option = module.params["option"]
    value = module.params["value"]
    values = module.params["values"]
    state = module.params["state"]
    exclusive = module.params["exclusive"]
    backup = module.params["backup"]
    no_extra_spaces = module.params["no_extra_spaces"]
    ignore_spaces = module.params["ignore_spaces"]
    allow_no_value = module.params["allow_no_value"]
    modify_inactive_option = module.params["modify_inactive_option"]
    create = module.params["create"]
    follow = module.params["follow"]

    if state == "present" and not allow_no_value and value is None and not values:
        module.fail_json(msg="Parameter 'value(s)' must be defined if state=present and allow_no_value=False.")

    if value is not None:
        values = [value]
    elif values is None:
        values = []

    if section_has_values:
        for condition in section_has_values:
            if condition["value"] is not None:
                condition["values"] = [condition["value"]]
            elif condition["values"] is None:
                condition["values"] = []
    #        raise Exception("section_has_values: {}".format(section_has_values))

    (changed, backup_file, diff, msg) = do_ini(
        module,
        path,
        section,
        section_has_values,
        option,
        values,
        state,
        exclusive,
        backup,
        no_extra_spaces,
        ignore_spaces,
        create,
        allow_no_value,
        modify_inactive_option,
        follow,
    )

    if not module.check_mode and os.path.exists(path):
        file_args = module.load_file_common_arguments(module.params)
        changed = module.set_fs_attributes_if_different(file_args, changed)

    results = dict(
        changed=changed,
        diff=diff,
        msg=msg,
        path=path,
    )
    if backup_file is not None:
        results["backup_file"] = backup_file

    # Mission complete
    module.exit_json(**results)


if __name__ == "__main__":
    main()



# This code is part of the Ansible collection community.docker, but is an independent component.
# This particular file, and this file only, is based on the Docker SDK for Python (https://github.com/docker/docker-py/)
#
# Copyright (c) 2016-2022 Docker, Inc.
#
# It is licensed under the Apache 2.0 license (see LICENSES/Apache-2.0.txt in this collection)
# SPDX-License-Identifier: Apache-2.0

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t

if t.TYPE_CHECKING:
    from subprocess import CalledProcessError


class StoreError(RuntimeError):
    pass


class CredentialsNotFound(StoreError):
    pass


class InitializationError(StoreError):
    pass


def process_store_error(cpe: CalledProcessError, program: str) -> StoreError:
    message = cpe.output.decode("utf-8")
    if "credentials not found in native keychain" in message:
        return CredentialsNotFound(f"No matching credentials in {program}")
    return StoreError(
        f'Credentials store {program} exited with "{cpe.output.decode("utf-8").strip()}".'
    )



# Copyright (c) 2018, Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
name: ksu
short_description: Kerberos substitute user
description:
  - This become plugins allows your remote/login user to execute commands as another user using the C(ksu) utility.
author: Ansible Core Team
options:
  become_user:
    description: User you 'become' to execute the task.
    type: string
    ini:
      - section: privilege_escalation
        key: become_user
      - section: ksu_become_plugin
        key: user
    vars:
      - name: ansible_become_user
      - name: ansible_ksu_user
    env:
      - name: ANSIBLE_BECOME_USER
      - name: ANSIBLE_KSU_USER
    required: true
  become_exe:
    description: C(ksu) executable.
    type: string
    default: ksu
    ini:
      - section: privilege_escalation
        key: become_exe
      - section: ksu_become_plugin
        key: executable
    vars:
      - name: ansible_become_exe
      - name: ansible_ksu_exe
    env:
      - name: ANSIBLE_BECOME_EXE
      - name: ANSIBLE_KSU_EXE
  become_flags:
    description: Options to pass to C(ksu).
    type: string
    default: ''
    ini:
      - section: privilege_escalation
        key: become_flags
      - section: ksu_become_plugin
        key: flags
    vars:
      - name: ansible_become_flags
      - name: ansible_ksu_flags
    env:
      - name: ANSIBLE_BECOME_FLAGS
      - name: ANSIBLE_KSU_FLAGS
  become_pass:
    description: C(ksu) password.
    type: string
    required: false
    vars:
      - name: ansible_ksu_pass
      - name: ansible_become_pass
      - name: ansible_become_password
    env:
      - name: ANSIBLE_BECOME_PASS
      - name: ANSIBLE_KSU_PASS
    ini:
      - section: ksu_become_plugin
        key: password
  prompt_l10n:
    description:
      - List of localized strings to match for prompt detection.
      - If empty the plugin uses the built-in one.
    type: list
    elements: string
    default: []
    ini:
      - section: ksu_become_plugin
        key: localized_prompts
    vars:
      - name: ansible_ksu_prompt_l10n
    env:
      - name: ANSIBLE_KSU_PROMPT_L10N
"""

import re

from ansible.module_utils.common.text.converters import to_bytes
from ansible.plugins.become import BecomeBase


class BecomeModule(BecomeBase):
    name = "community.general.ksu"

    # messages for detecting prompted password issues
    fail = ("Password incorrect",)
    missing = ("No password given",)

    def check_password_prompt(self, b_output):
        """checks if the expected password prompt exists in b_output"""

        prompts = self.get_option("prompt_l10n") or ["Kerberos password for .*@.*:"]
        b_prompt = b"|".join(to_bytes(p) for p in prompts)

        return bool(re.match(b_prompt, b_output))

    def build_become_command(self, cmd, shell):
        super().build_become_command(cmd, shell)

        # Prompt handling for ``ksu`` is more complicated, this
        # is used to satisfy the connection plugin
        self.prompt = True

        if not cmd:
            return cmd

        exe = self.get_option("become_exe")

        flags = self.get_option("become_flags")
        user = self.get_option("become_user")
        return f"{exe} {user} {flags} -e {self._build_success_command(cmd, shell)} "



# Copyright (c) 2016, Dag Wieers <dag@wieers.com>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
name: dense
type: stdout
short_description: Minimal stdout output
extends_documentation_fragment:
  - ansible.builtin.default_callback
description:
  - When in verbose mode it acts the same as the default callback.
author:
  - Dag Wieers (@dagwieers)
requirements:
  - set as stdout in configuration
"""

HAS_OD = False
try:
    from collections import OrderedDict

    HAS_OD = True
except ImportError:
    pass

import sys
from collections.abc import MutableMapping, MutableSequence

from ansible.plugins.callback.default import CallbackModule as CallbackModule_default
from ansible.utils.color import colorize, hostcolor
from ansible.utils.display import Display

display = Display()


# Design goals:
#
#  + On screen there should only be relevant stuff
#    - How far are we ? (during run, last line)
#    - What issues occurred
#    - What changes occurred
#    - Diff output (in diff-mode)
#
#  + If verbosity increases, act as default output
#    So that users can easily switch to default for troubleshooting
#
#  + Rewrite the output during processing
#    - We use the cursor to indicate where in the task we are.
#      Output after the prompt is the output of the previous task.
#    - If we would clear the line at the start of a task, there would often
#      be no information at all, so we leave it until it gets updated
#
#  + Use the same color-conventions of Ansible
#
#  + Ensure the verbose output (-v) is also dense.
#    Remove information that is not essential (eg. timestamps, status)


# TODO:
#
#  + Properly test for terminal capabilities, and fall back to default
#  + Modify Ansible mechanism so we don't need to use sys.stdout directly
#  + Find an elegant solution for progress bar line wrapping


# FIXME: Importing constants as C simply does not work, beats me :-/
# from ansible import constants as C
class C:
    COLOR_HIGHLIGHT = "white"
    COLOR_VERBOSE = "blue"
    COLOR_WARN = "bright purple"
    COLOR_ERROR = "red"
    COLOR_DEBUG = "dark gray"
    COLOR_DEPRECATE = "purple"
    COLOR_SKIP = "cyan"
    COLOR_UNREACHABLE = "bright red"
    COLOR_OK = "green"
    COLOR_CHANGED = "yellow"


# Taken from Dstat
class vt100:
    black = "\033[0;30m"
    darkred = "\033[0;31m"
    darkgreen = "\033[0;32m"
    darkyellow = "\033[0;33m"
    darkblue = "\033[0;34m"
    darkmagenta = "\033[0;35m"
    darkcyan = "\033[0;36m"
    gray = "\033[0;37m"

    darkgray = "\033[1;30m"
    red = "\033[1;31m"
    green = "\033[1;32m"
    yellow = "\033[1;33m"
    blue = "\033[1;34m"
    magenta = "\033[1;35m"
    cyan = "\033[1;36m"
    white = "\033[1;37m"

    blackbg = "\033[40m"
    redbg = "\033[41m"
    greenbg = "\033[42m"
    yellowbg = "\033[43m"
    bluebg = "\033[44m"
    magentabg = "\033[45m"
    cyanbg = "\033[46m"
    whitebg = "\033[47m"

    reset = "\033[0;0m"
    bold = "\033[1m"
    reverse = "\033[2m"
    underline = "\033[4m"

    clear = "\033[2J"
    #    clearline = '\033[K'
    clearline = "\033[2K"
    save = "\033[s"
    restore = "\033[u"
    save_all = "\0337"
    restore_all = "\0338"
    linewrap = "\033[7h"
    nolinewrap = "\033[7l"

    up = "\033[1A"
    down = "\033[1B"
    right = "\033[1C"
    left = "\033[1D"


colors = dict(
    ok=vt100.darkgreen,
    changed=vt100.darkyellow,
    skipped=vt100.darkcyan,
    ignored=vt100.cyanbg + vt100.red,
    failed=vt100.darkred,
    unreachable=vt100.red,
)

states = ("skipped", "ok", "changed", "failed", "unreachable")


class CallbackModule(CallbackModule_default):
    """
    This is the dense callback interface, where screen estate is still valued.
    """

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = "stdout"
    CALLBACK_NAME = "dense"

    def __init__(self):
        # From CallbackModule
        self._display = display

        if HAS_OD:
            self.disabled = False
            self.super_ref = super()
            self.super_ref.__init__()

            # Attributes to remove from results for more density
            self.removed_attributes = (
                #                'changed',
                "delta",
                #                'diff',
                "end",
                "failed",
                "failed_when_result",
                "invocation",
                "start",
                "stdout_lines",
            )

            # Initiate data structures
            self.hosts = OrderedDict()
            self.keep = False
            self.shown_title = False
            self.count = dict(play=0, handler=0, task=0)
            self.type = "foo"

            # Start immediately on the first line
            sys.stdout.write(vt100.reset + vt100.save + vt100.clearline)
            sys.stdout.flush()
        else:
            display.warning(
                "The 'dense' callback plugin requires OrderedDict which is not available in this version of python, disabling."
            )
            self.disabled = True

    def __del__(self):
        sys.stdout.write(f"{vt100.restore}{vt100.reset}\n{vt100.save}{vt100.clearline}")

    def _add_host(self, result, status):
        name = result._host.get_name()

        # Add a new status in case a failed task is ignored
        if status == "failed" and result._task.ignore_errors:
            status = "ignored"

        # Check if we have to update an existing state (when looping over items)
        if name not in self.hosts:
            self.hosts[name] = dict(state=status)
        elif states.index(self.hosts[name]["state"]) < states.index(status):
            self.hosts[name]["state"] = status

        # Store delegated hostname, if needed
        delegated_vars = result._result.get("_ansible_delegated_vars", None)
        if delegated_vars:
            self.hosts[name]["delegate"] = delegated_vars["ansible_host"]

        # Print progress bar
        self._display_progress(result)

        #        # Ensure that tasks with changes/failures stay on-screen, and during diff-mode
        #        if status in ['changed', 'failed', 'unreachable'] or (result.get('_diff_mode', False) and result._resultget('diff', False)):
        # Ensure that tasks with changes/failures stay on-screen
        if status in ["changed", "failed", "unreachable"]:
            self.keep = True

            if self._display.verbosity == 1:
                # Print task title, if needed
                self._display_task_banner()
                self._display_results(result, status)

    def _clean_results(self, result):
        # Remove non-essential attributes
        for attr in self.removed_attributes:
            if attr in result:
                del result[attr]

        # Remove empty attributes (list, dict, str)
        for attr in result.copy():
            if isinstance(result[attr], (MutableSequence, MutableMapping, bytes, str)):
                if not result[attr]:
                    del result[attr]

    def _handle_exceptions(self, result):
        if "exception" in result:
            # Remove the exception from the result so it is not shown every time
            del result["exception"]

            if self._display.verbosity == 1:
                return "An exception occurred during task execution. To see the full traceback, use -vvv."

    def _display_progress(self, result=None):
        # Always rewrite the complete line
        sys.stdout.write(vt100.restore + vt100.reset + vt100.clearline + vt100.nolinewrap + vt100.underline)
        sys.stdout.write(f"{self.type} {self.count[self.type]}:")
        sys.stdout.write(vt100.reset)
        sys.stdout.flush()

        # Print out each host in its own status-color
        for name in self.hosts:
            sys.stdout.write(" ")
            if self.hosts[name].get("delegate", None):
                sys.stdout.write(f"{self.hosts[name]['delegate']}>")
            sys.stdout.write(colors[self.hosts[name]["state"]] + name + vt100.reset)
            sys.stdout.flush()

        sys.stdout.write(vt100.linewrap)

    def _display_task_banner(self):
        if not self.shown_title:
            self.shown_title = True
            sys.stdout.write(vt100.restore + vt100.reset + vt100.clearline + vt100.underline)
            sys.stdout.write(f"{self.type} {self.count[self.type]}: {self.task.get_name().strip()}")
            sys.stdout.write(f"{vt100.restore}{vt100.reset}\n{vt100.save}{vt100.clearline}")
            sys.stdout.flush()
        else:
            sys.stdout.write(vt100.restore + vt100.reset + vt100.clearline)
        self.keep = False

    def _display_results(self, result, status):
        # Leave the previous task on screen (as it has changes/errors)
        if self._display.verbosity == 0 and self.keep:
            sys.stdout.write(f"{vt100.restore}{vt100.reset}\n{vt100.save}{vt100.clearline}")
        else:
            sys.stdout.write(vt100.restore + vt100.reset + vt100.clearline)
        self.keep = False

        self._clean_results(result._result)

        dump = ""
        if result._task.action == "include":
            return
        elif status == "ok":
            return
        elif status == "ignored":
            dump = self._handle_exceptions(result._result)
        elif status == "failed":
            dump = self._handle_exceptions(result._result)
        elif status == "unreachable":
            dump = result._result["msg"]

        if not dump:
            dump = self._dump_results(result._result)

        if result._task.loop and "results" in result._result:
            self._process_items(result)
        else:
            sys.stdout.write(f"{colors[status] + status}: ")

            delegated_vars = result._result.get("_ansible_delegated_vars", None)
            if delegated_vars:
                sys.stdout.write(
                    f"{vt100.reset}{result._host.get_name()}>{colors[status]}{delegated_vars['ansible_host']}"
                )
            else:
                sys.stdout.write(result._host.get_name())

            sys.stdout.write(f": {dump}\n")
            sys.stdout.write(f"{vt100.reset}{vt100.save}{vt100.clearline}")
            sys.stdout.flush()

        if status == "changed":
            self._handle_warnings(result._result)

    def v2_playbook_on_play_start(self, play):
        # Leave the previous task on screen (as it has changes/errors)
        if self._display.verbosity == 0 and self.keep:
            sys.stdout.write(f"{vt100.restore}{vt100.reset}\n{vt100.save}{vt100.clearline}{vt100.bold}")
        else:
            sys.stdout.write(vt100.restore + vt100.reset + vt100.clearline + vt100.bold)

        # Reset at the start of each play
        self.keep = False
        self.count.update(dict(handler=0, task=0))
        self.count["play"] += 1
        self.play = play

        # Write the next play on screen IN UPPERCASE, and make it permanent
        name = play.get_name().strip()
        if not name:
            name = "unnamed"
        sys.stdout.write(f"PLAY {self.count['play']}: {name.upper()}")
        sys.stdout.write(f"{vt100.restore}{vt100.reset}\n{vt100.save}{vt100.clearline}")
        sys.stdout.flush()

    def v2_playbook_on_task_start(self, task, is_conditional):
        # Leave the previous task on screen (as it has changes/errors)
        if self._display.verbosity == 0 and self.keep:
            sys.stdout.write(f"{vt100.restore}{vt100.reset}\n{vt100.save}{vt100.clearline}{vt100.underline}")
        else:
            # Do not clear line, since we want to retain the previous output
            sys.stdout.write(vt100.restore + vt100.reset + vt100.underline)

        # Reset at the start of each task
        self.keep = False
        self.shown_title = False
        self.hosts = OrderedDict()
        self.task = task
        self.type = "task"

        # Enumerate task if not setup (task names are too long for dense output)
        if task.get_name() != "setup":
            self.count["task"] += 1

        # Write the next task on screen (behind the prompt is the previous output)
        sys.stdout.write(f"{self.type} {self.count[self.type]}.")
        sys.stdout.write(vt100.reset)
        sys.stdout.flush()

    def v2_playbook_on_handler_task_start(self, task):
        # Leave the previous task on screen (as it has changes/errors)
        if self._display.verbosity == 0 and self.keep:
            sys.stdout.write(f"{vt100.restore}{vt100.reset}\n{vt100.save}{vt100.clearline}{vt100.underline}")
        else:
            sys.stdout.write(vt100.restore + vt100.reset + vt100.clearline + vt100.underline)

        # Reset at the start of each handler
        self.keep = False
        self.shown_title = False
        self.hosts = OrderedDict()
        self.task = task
        self.type = "handler"

        # Enumerate handler if not setup (handler names may be too long for dense output)
        if task.get_name() != "setup":
            self.count[self.type] += 1

        # Write the next task on screen (behind the prompt is the previous output)
        sys.stdout.write(f"{self.type} {self.count[self.type]}.")
        sys.stdout.write(vt100.reset)
        sys.stdout.flush()

    def v2_playbook_on_cleanup_task_start(self, task):
        # TBD
        sys.stdout.write("cleanup.")
        sys.stdout.flush()

    def v2_runner_on_failed(self, result, ignore_errors=False):
        self._add_host(result, "failed")

    def v2_runner_on_ok(self, result):
        if result._result.get("changed", False):
            self._add_host(result, "changed")
        else:
            self._add_host(result, "ok")

    def v2_runner_on_skipped(self, result):
        self._add_host(result, "skipped")

    def v2_runner_on_unreachable(self, result):
        self._add_host(result, "unreachable")

    def v2_runner_on_include(self, included_file):
        pass

    def v2_runner_on_file_diff(self, result, diff):
        sys.stdout.write(vt100.bold)
        self.super_ref.v2_runner_on_file_diff(result, diff)
        sys.stdout.write(vt100.reset)

    def v2_on_file_diff(self, result):
        sys.stdout.write(vt100.bold)
        self.super_ref.v2_on_file_diff(result)
        sys.stdout.write(vt100.reset)

    # Old definition in v2.0
    def v2_playbook_item_on_ok(self, result):
        self.v2_runner_item_on_ok(result)

    def v2_runner_item_on_ok(self, result):
        if result._result.get("changed", False):
            self._add_host(result, "changed")
        else:
            self._add_host(result, "ok")

    # Old definition in v2.0
    def v2_playbook_item_on_failed(self, result):
        self.v2_runner_item_on_failed(result)

    def v2_runner_item_on_failed(self, result):
        self._add_host(result, "failed")

    # Old definition in v2.0
    def v2_playbook_item_on_skipped(self, result):
        self.v2_runner_item_on_skipped(result)

    def v2_runner_item_on_skipped(self, result):
        self._add_host(result, "skipped")

    def v2_playbook_on_no_hosts_remaining(self):
        if self._display.verbosity == 0 and self.keep:
            sys.stdout.write(f"{vt100.restore}{vt100.reset}\n{vt100.save}{vt100.clearline}")
        else:
            sys.stdout.write(vt100.restore + vt100.reset + vt100.clearline)
        self.keep = False

        sys.stdout.write(f"{vt100.white + vt100.redbg}NO MORE HOSTS LEFT")
        sys.stdout.write(f"{vt100.restore}{vt100.reset}\n{vt100.save}{vt100.clearline}")
        sys.stdout.flush()

    def v2_playbook_on_include(self, included_file):
        pass

    def v2_playbook_on_stats(self, stats):
        if self._display.verbosity == 0 and self.keep:
            sys.stdout.write(f"{vt100.restore}{vt100.reset}\n{vt100.save}{vt100.clearline}")
        else:
            sys.stdout.write(vt100.restore + vt100.reset + vt100.clearline)

        # In normal mode screen output should be sufficient, summary is redundant
        if self._display.verbosity == 0:
            return

        sys.stdout.write(vt100.bold + vt100.underline)
        sys.stdout.write("SUMMARY")

        sys.stdout.write(f"{vt100.restore}{vt100.reset}\n{vt100.save}{vt100.clearline}")
        sys.stdout.flush()

        hosts = sorted(stats.processed.keys())
        for h in hosts:
            t = stats.summarize(h)
            self._display.display(
                f"{hostcolor(h, t)} : {colorize('ok', t['ok'], C.COLOR_OK)} {colorize('changed', t['changed'], C.COLOR_CHANGED)} "
                f"{colorize('unreachable', t['unreachable'], C.COLOR_UNREACHABLE)} {colorize('failed', t['failures'], C.COLOR_ERROR)} "
                f"{colorize('rescued', t['rescued'], C.COLOR_OK)} {colorize('ignored', t['ignored'], C.COLOR_WARN)}",
                screen_only=True,
            )


# When using -vv or higher, simply do the default action
if display.verbosity >= 2 or not HAS_OD:
    CallbackModule = CallbackModule_default  # type: ignore



#!/usr/bin/python

# Copyright (c) 2022, Guillaume MARTINEZ (lunik@tiwabbit.fr)
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: gitlab_project_badge
short_description: Manage project badges on GitLab Server
version_added: 6.1.0
description:
  - This module allows to add and remove badges to/from a project.
author: Guillaume MARTINEZ (@Lunik)
requirements:
  - C(owner) or C(maintainer) rights to project on the GitLab server
extends_documentation_fragment:
  - community.general._auth_basic
  - community.general._gitlab
  - community.general._attributes

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none

options:
  project:
    description:
      - The name (or full path) of the GitLab project the badge is added to/removed from.
    required: true
    type: str

  state:
    description:
      - State of the badge in the project.
      - On V(present), it adds a badge to a GitLab project.
      - On V(absent), it removes a badge from a GitLab project.
    choices: ['present', 'absent']
    default: 'present'
    type: str

  link_url:
    description:
      - The URL associated with the badge.
    required: true
    type: str

  image_url:
    description:
      - The image URL of the badge.
      - A badge is identified by this URL.
    required: true
    type: str
"""

EXAMPLES = r"""
- name: Add a badge to a GitLab Project
  community.general.gitlab_project_badge:
    api_url: 'https://example.gitlab.com'
    api_token: 'Your-Private-Token'
    project: projectname
    state: present
    link_url: 'https://example.gitlab.com/%{project_path}'
    image_url: 'https://example.gitlab.com/%{project_path}/badges/%{default_branch}/pipeline.svg'

- name: Remove a badge from a GitLab Project
  community.general.gitlab_project_badge:
    api_url: 'https://example.gitlab.com'
    api_token: 'Your-Private-Token'
    project: projectname
    state: absent
    link_url: 'https://example.gitlab.com/%{project_path}'
    image_url: 'https://example.gitlab.com/%{project_path}/badges/%{default_branch}/pipeline.svg'
"""

RETURN = r"""
badge:
  description: The badge information.
  returned: when O(state=present)
  type: dict
  sample:
    id: 1
    link_url: 'http://example.com/ci_status.svg?project=%{project_path}&ref=%{default_branch}'
    image_url: 'https://shields.io/my/badge'
    rendered_link_url: 'http://example.com/ci_status.svg?project=example-org/example-project&ref=master'
    rendered_image_url: 'https://shields.io/my/badge'
    kind: project
"""

from ansible.module_utils.api import basic_auth_argument_spec
from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._gitlab import (
    auth_argument_spec,
    find_project,
    gitlab_authentication,
    list_all_kwargs,
)


def present_strategy(module, gl, project, wished_badge):
    changed = False

    existing_badge = None
    for badge in project.badges.list(**list_all_kwargs):
        if badge.image_url == wished_badge["image_url"]:
            existing_badge = badge
            break

    if not existing_badge:
        changed = True
        if module.check_mode:
            return changed, {"status": "A project badge would be created."}

        badge = project.badges.create(wished_badge)
        return changed, badge.attributes

    if existing_badge.link_url != wished_badge["link_url"]:
        changed = True
        existing_badge.link_url = wished_badge["link_url"]

    if changed:
        if module.check_mode:
            return changed, {"status": "Project badge attributes would be changed."}

        existing_badge.save()

    return changed, existing_badge.attributes


def absent_strategy(module, gl, project, wished_badge):
    changed = False

    existing_badge = None
    for badge in project.badges.list(**list_all_kwargs):
        if badge.image_url == wished_badge["image_url"]:
            existing_badge = badge
            break

    if not existing_badge:
        return changed, None

    changed = True
    if module.check_mode:
        return changed, {"status": "Project badge would be destroyed."}

    existing_badge.delete()

    return changed, None


state_strategy = {"present": present_strategy, "absent": absent_strategy}


def core(module):
    # check prerequisites and connect to gitlab server
    gl = gitlab_authentication(module)

    gitlab_project = module.params["project"]
    state = module.params["state"]

    project = find_project(gl, gitlab_project)
    # project doesn't exist
    if not project:
        module.fail_json(msg=f"project '{gitlab_project}' not found.")

    wished_badge = {
        "link_url": module.params["link_url"],
        "image_url": module.params["image_url"],
    }

    changed, summary = state_strategy[state](module=module, gl=gl, project=project, wished_badge=wished_badge)

    module.exit_json(changed=changed, badge=summary)


def main():
    argument_spec = basic_auth_argument_spec()
    argument_spec.update(auth_argument_spec())
    argument_spec.update(
        dict(
            project=dict(type="str", required=True),
            state=dict(type="str", default="present", choices=["present", "absent"]),
            link_url=dict(type="str", required=True),
            image_url=dict(type="str", required=True),
        )
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=[
            ["api_username", "api_token"],
            ["api_username", "api_oauth_token"],
            ["api_username", "api_job_token"],
            ["api_token", "api_oauth_token"],
            ["api_token", "api_job_token"],
        ],
        required_together=[
            ["api_username", "api_password"],
        ],
        required_one_of=[
            ["api_username", "api_token", "api_oauth_token", "api_job_token"],
        ],
        supports_check_mode=True,
    )

    core(module)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Red Hat, Inc.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import json
import string
import typing

if typing.TYPE_CHECKING:
    from typing import Any
    from typing import Dict
    from typing import Generator
    from typing import List
    from typing import Optional
    from typing import Tuple

    from .botocore import ClientType
    from .modules import AnsibleAWSModule

from urllib.parse import urlparse

try:
    from hashlib import md5

    HAS_MD5 = True
except ImportError:
    HAS_MD5 = False

try:
    import botocore
except ImportError:
    pass  # Handled by the calling module


from ansible.module_utils.basic import to_text

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry
from ansible_collections.amazon.aws.plugins.module_utils.tagging import ansible_dict_to_boto3_tag_list
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict

IGNORE_S3_DROP_IN_EXCEPTIONS = ["XNotImplemented", "NotImplemented", "AccessControlListNotSupported"]

from ._s3 import common as _common
from ._s3 import transformations as _transformations
from ._s3 import waiters as _waiters

S3ErrorHandler = _common.S3ErrorHandler

AnsibleS3Error = _common.AnsibleS3Error
AnsibleS3Sigv4RequiredError = _common.AnsibleS3Error
AnsibleS3PermissionsError = _common.AnsibleS3PermissionsError
AnsibleS3SupportError = _common.AnsibleS3SupportError
AnsibleS3ACLSupportError = _common.AnsibleS3ACLSupportError
AnsibleS3RegionSupportError = _common.AnsibleS3RegionSupportError

normalize_s3_bucket_versioning = _transformations.normalize_s3_bucket_versioning
normalize_s3_bucket_public_access = _transformations.normalize_s3_bucket_public_access
normalize_s3_bucket_acls = _transformations.normalize_s3_bucket_acls
normalize_s3_bucket_logging = _transformations.normalize_s3_bucket_logging
s3_acl_to_name = _transformations.s3_acl_to_name
merge_tags = _transformations.merge_tags


# ========================================
# S3 Client Wrappers - Bucket Operations
# ========================================


def get_s3_waiter(client: ClientType, waiter_name: str) -> Any:
    return _waiters.waiter_factory.get_waiter(client, waiter_name)


@S3ErrorHandler.list_error_handler("get bucket accelerate configuration", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_bucket_accelerate_configuration(client: ClientType, bucket_name: str) -> Dict:
    """
    Get transfer accelerate configuration of the S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        Transfer accelerate status of the S3 bucket.
    """
    try:
        return client.get_bucket_accelerate_configuration(Bucket=bucket_name)
    except is_boto3_error_code(["UnsupportedArgument", "MethodNotAllowed"]) as e:
        # aws-gov throws UnsupportedArgument (consistently)
        # aws throws MethodNotAllowed where acceleration isn't available /yet/
        raise AnsibleS3RegionSupportError(
            message="Failed to get bucket accelerate configuration (not available in S3 bucket region)", exception=e
        ) from e


@S3ErrorHandler.list_error_handler("get bucket ACLs", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_bucket_acl(client: ClientType, bucket_name: str) -> Dict:
    """
    Get ACLs of the S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        ACL configuration of the S3 bucket.
    """
    return client.get_bucket_acl(Bucket=bucket_name)


@S3ErrorHandler.list_error_handler("get bucket encryption settings", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_bucket_encryption(client: ClientType, bucket_name: str) -> Dict:
    """
    Retrieve the encryption configuration for an S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        Encryption configuration of the bucket.
    """
    return client.get_bucket_encryption(Bucket=bucket_name).get("ServerSideEncryptionConfiguration")


@S3ErrorHandler.list_error_handler("get bucket ownership settings", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_bucket_ownership_controls(client: ClientType, bucket_name: str) -> Dict:
    """
    Retrieve bucket ownership controls for S3 bucket
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        Bucket ownership controls of the bucket.
    """
    return client.get_bucket_ownership_controls(Bucket=bucket_name).get("OwnershipControls")


@S3ErrorHandler.list_error_handler("get bucket policy", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_bucket_policy(client: ClientType, bucket_name: str) -> Dict:
    """
    Retrieve bucket policy for an S3 bucket
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        Bucket policy controls of the bucket.
    """
    policy = client.get_bucket_policy(Bucket=bucket_name).get("Policy")
    try:
        return json.loads(policy)
    except ValueError as e:
        raise AnsibleS3Error(exception=e, message="Unable to parse current bucket policy") from e


@S3ErrorHandler.list_error_handler("get bucket public access block settings", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_bucket_public_access_block(client: ClientType, bucket_name: str) -> Dict:
    """
    Get current public access block configuration for a bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        The current public access block configuration for the bucket.
    """
    return client.get_public_access_block(Bucket=bucket_name).get("PublicAccessBlockConfiguration")


@S3ErrorHandler.list_error_handler("get bucket request payment settings", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_bucket_request_payment(client: ClientType, bucket_name: str) -> Dict:
    """
    Retrieve bucket request payment settings for an S3 bucket
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        Request payment settings of the bucket.
    """
    return client.get_bucket_request_payment(Bucket=bucket_name).get("Payer")


@S3ErrorHandler.list_error_handler("get bucket tags", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_bucket_tagging(client: ClientType, bucket_name: str) -> Dict:
    """
    Retrieve tagging for an S3 bucket
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        The current tags dictionary applied to the bucket.
    """
    current_tags = client.get_bucket_tagging(Bucket=bucket_name).get("TagSet")
    return boto3_tag_list_to_ansible_dict(current_tags)


@S3ErrorHandler.list_error_handler("get bucket versioning settings", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_bucket_versioning(client: ClientType, bucket_name: str) -> Dict:
    """
    Retrieve bucket versioning settings for an S3 bucket
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        Versioning settings of the bucket.
    """
    return client.get_bucket_versioning(Bucket=bucket_name)


@S3ErrorHandler.list_error_handler("get bucket logging configuration", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_bucket_logging(client: ClientType, bucket_name: str) -> Dict:
    """
    Retrieve bucket logging configuration for an S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        Logging configuration of the bucket.
    """
    return client.get_bucket_logging(Bucket=bucket_name)


@S3ErrorHandler.list_error_handler("get bucket object lock settings", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_object_lock_configuration(client: ClientType, bucket_name: str) -> Dict:
    """
    Retrieve object lock configuration for an S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        The object lock configuration of the bucket.
    """
    return client.get_object_lock_configuration(Bucket=bucket_name).get("ObjectLockConfiguration")


@S3ErrorHandler.list_error_handler("determine if bucket exists", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["OperationAborted"])
def head_s3_bucket(client: ClientType, bucket_name: str) -> Dict:
    """
    Retrieve basic information about an S3 bucket
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        Basic information about the bucket.
    """
    return client.head_bucket(Bucket=bucket_name)


@S3ErrorHandler.list_error_handler("get bucket location", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_bucket_location(client: ClientType, bucket_name: str) -> Dict:
    """
    Retrieve the AWS region where an S3 bucket is located.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.

    Returns:
        Dict: Bucket location information containing 'LocationConstraint'.
              Returns {} if bucket doesn't exist (404).
              Note: LocationConstraint is None for us-east-1 buckets.

    Raises:
        AnsibleS3PermissionsError: If access is denied (403).
        AnsibleS3Error: For other S3 errors.
    """
    return client.get_bucket_location(Bucket=bucket_name)


@AWSRetry.jittered_backoff()
def s3_bucket_exists(client: ClientType, bucket_name: str) -> bool:
    """
    Check if an S3 bucket exists.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.

    Returns:
        bool: True if bucket exists, False otherwise.

    Raises:
        AnsibleS3PermissionsError: If access is denied (403).
        AnsibleS3Error: For other S3 errors.
    """
    result = head_s3_bucket(client, bucket_name)
    return bool(result)


@S3ErrorHandler.list_error_handler("list buckets", [])
@AWSRetry.jittered_backoff()
def list_s3_buckets(client: ClientType) -> List[Dict]:
    """
    List all S3 buckets in the account.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.

    Returns:
        List[Dict]: List of bucket information dictionaries containing 'Name' and 'CreationDate'.
                    Returns [] on error.

    Raises:
        AnsibleS3PermissionsError: If access is denied (403).
        AnsibleS3Error: For other S3 errors.
    """
    return client.list_buckets().get("Buckets", [])


@S3ErrorHandler.list_error_handler("get bucket inventory settings", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def _list_bucket_inventory_configurations(client: ClientType, **params) -> Dict:
    return client.list_bucket_inventory_configurations(**params)


# _list_bucket_inventory_configurations is a workaround for a missing paginator for listing
# bucket inventory configuration in boto3:
# https://github.com/boto/botocore/blob/1.34.141/botocore/data/s3/2006-03-01/paginators-1.json
def list_bucket_inventory_configurations(client: ClientType, bucket_name: str) -> list:
    first_iteration_completed = False
    next_token = None

    response = _list_bucket_inventory_configurations(client, Bucket=bucket_name)
    next_token = response.get("NextToken", None)

    if next_token is None:
        return response.get("InventoryConfigurationList", [])

    entries = []
    while next_token is not None:
        if first_iteration_completed:
            response = _list_bucket_inventory_configurations(client, NextToken=next_token, Bucket=bucket_name)
        first_iteration_completed = True
        entries.extend(response["InventoryConfigurationList"])
        next_token = response.get("NextToken")
    return entries


@S3ErrorHandler.common_error_handler("set object lock configuration")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_s3_object_lock_configuration(client: ClientType, bucket_name: str, object_lock_default_retention: str) -> None:
    """
    Set object lock configuration for an S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_lock_default_retention (dict): A dictionary containing the object
        lock default retention configuration to be set on the bucket.
    Returns:
        None
    """
    conf = {"ObjectLockEnabled": "Enabled", "Rule": {"DefaultRetention": object_lock_default_retention}}
    client.put_object_lock_configuration(Bucket=bucket_name, ObjectLockConfiguration=conf)


@S3ErrorHandler.common_error_handler("set bucket acceleration configuration")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_s3_bucket_accelerate_configuration(client: ClientType, bucket_name: str) -> None:
    """
    Enable transfer accelerate for the S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        None
    """
    client.put_bucket_accelerate_configuration(Bucket=bucket_name, AccelerateConfiguration={"Status": "Enabled"})


@S3ErrorHandler.common_error_handler("set bucket inventory configuration")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_s3_bucket_inventory(client: ClientType, bucket_name: str, inventory: dict) -> None:
    """
    Set inventory settings for an S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        inventory (dict): A dictionary containing the inventory settings to be set on the bucket.
    Returns:
        None
    """
    try:
        client.put_bucket_inventory_configuration(
            Bucket=bucket_name, InventoryConfiguration=inventory, Id=inventory.get("Id")
        )
    except is_boto3_error_code("InvalidS3DestinationBucket") as e:
        raise AnsibleS3Error("Invalid destination bucket ARN") from e


@S3ErrorHandler.common_error_handler("set bucket tagging")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_s3_bucket_tagging(client: ClientType, bucket_name: str, tags: dict) -> None:
    """
    Set tags for an S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        tags (dict): A dictionary containing the tags to be set on the bucket.
    Returns:
        None
    """
    client.put_bucket_tagging(Bucket=bucket_name, Tagging={"TagSet": ansible_dict_to_boto3_tag_list(tags)})


@S3ErrorHandler.common_error_handler("set bucket policy")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_s3_bucket_policy(client: ClientType, bucket_name: str, policy: dict) -> None:
    """
    Set the policy for an S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        policy (dict): A dictionary containing the policy to be set on the bucket.
    Returns:
        None
    """
    client.put_bucket_policy(Bucket=bucket_name, Policy=json.dumps(policy))


@S3ErrorHandler.common_error_handler("set bucket request payment configuration")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_s3_bucket_request_payment(client: ClientType, bucket_name: str, payer: str) -> None:
    """
    Set the request payment configuration for an S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        payer (str): The entity responsible for charges related to fulfilling the request.
    Returns:
        None
    """
    client.put_bucket_request_payment(Bucket=bucket_name, RequestPaymentConfiguration={"Payer": payer})


@S3ErrorHandler.common_error_handler("set bucket versioning configuration")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_s3_bucket_versioning(client: ClientType, bucket_name: str, required_versioning: str) -> None:
    """
    Set the versioning configuration for an S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        required_versioning (str): The desired versioning state for the bucket ("Enabled", "Suspended").
    Returns:
        None
    """
    client.put_bucket_versioning(Bucket=bucket_name, VersioningConfiguration={"Status": required_versioning})


@S3ErrorHandler.common_error_handler("set bucket logging configuration")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_s3_bucket_logging(client: ClientType, bucket_name: str, logging_config: dict) -> None:
    """
    Set the logging configuration for an S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        logging_config (dict): The logging configuration to set.
    Returns:
        None
    """
    client.put_bucket_logging(Bucket=bucket_name, BucketLoggingStatus=logging_config)


@S3ErrorHandler.common_error_handler("set bucket encryption")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_s3_bucket_encryption(client: ClientType, bucket_name: str, encryption: dict) -> None:
    """
    Set the encryption configuration for an S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        encryption (dict): A dictionary containing the encryption configuration.
    Returns:
        None
    """
    server_side_encryption_configuration = {"Rules": [{"ApplyServerSideEncryptionByDefault": encryption}]}
    client.put_bucket_encryption(
        Bucket=bucket_name, ServerSideEncryptionConfiguration=server_side_encryption_configuration
    )


@S3ErrorHandler.common_error_handler("set public access block configuration")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_s3_bucket_public_access(client: ClientType, bucket_name: str, public_access_config: dict) -> None:
    """
    Put new public access block to S3 bucket
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        public_access_config (dict): The public access block configuration.
    Returns:
        None
    """
    client.put_public_access_block(Bucket=bucket_name, PublicAccessBlockConfiguration=public_access_config)


@S3ErrorHandler.common_error_handler("set bucket ACL")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_s3_bucket_acl(client: ClientType, bucket_name: str, acl: str) -> None:
    """
    Applies a canned ACL to an S3 bucket
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        acl (str): The ACL
    Returns:
        None
    """
    client.put_bucket_acl(Bucket=bucket_name, ACL=acl)


@S3ErrorHandler.common_error_handler("set bucket ownership controls")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_s3_bucket_ownership(client: ClientType, bucket_name: str, object_ownership: str) -> None:
    """
    Put bucket ownership controls for S3 bucket
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_ownership (str): The object ownership control setting.
    Returns:
        None
    """
    client.put_bucket_ownership_controls(
        Bucket=bucket_name, OwnershipControls={"Rules": [{"ObjectOwnership": object_ownership}]}
    )


@S3ErrorHandler.deletion_error_handler("set bucket acceleration configuration")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def delete_s3_bucket_accelerate_configuration(client: ClientType, bucket_name: str) -> None:
    """
    Disable transfer accelerate for the S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        None
    """
    client.put_bucket_accelerate_configuration(Bucket=bucket_name, AccelerateConfiguration={"Status": "Suspended"})


@S3ErrorHandler.deletion_error_handler("delete bucket inventory configuration")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def delete_s3_bucket_inventory(client: ClientType, bucket_name: str, inventory_id: str) -> None:
    """
    Delete the inventory settings for an S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        inventory_id (str): The ID used to identify the inventory configuration
    Returns:
        None
    """
    client.delete_bucket_inventory_configuration(Bucket=bucket_name, Id=inventory_id)


@S3ErrorHandler.deletion_error_handler("delete bucket tagging")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def delete_s3_bucket_tagging(client: ClientType, bucket_name: str) -> None:
    """
    Delete the tagging configuration of an S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        None
    """
    client.delete_bucket_tagging(Bucket=bucket_name)


@S3ErrorHandler.deletion_error_handler("delete bucket encryption")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def delete_s3_bucket_encryption(client: ClientType, bucket_name: str) -> None:
    """
    Delete the encryption configuration of an S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        None
    """
    client.delete_bucket_encryption(Bucket=bucket_name)


@S3ErrorHandler.deletion_error_handler("delete bucket")
@AWSRetry.jittered_backoff(max_delay=240, catch_extra_error_codes=["OperationAborted"])
def delete_s3_bucket(client: ClientType, bucket_name: str) -> None:
    """
    Delete an S3 bucket.
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        None
    """
    client.delete_bucket(Bucket=bucket_name)


@S3ErrorHandler.deletion_error_handler("delete bucket policy")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def delete_s3_bucket_policy(client: ClientType, bucket_name: str) -> None:
    """
    Delete policy from S3 bucket
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        None
    """
    client.delete_bucket_policy(Bucket=bucket_name)


@S3ErrorHandler.deletion_error_handler("delete public access block configuration")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def delete_s3_bucket_public_access(client: ClientType, bucket_name: str) -> None:
    """
    Delete public access block from S3 bucket
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        None
    """
    client.delete_public_access_block(Bucket=bucket_name)


@S3ErrorHandler.deletion_error_handler("delete bucket ownership controls")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def delete_s3_bucket_ownership(client: ClientType, bucket_name: str) -> None:
    """
    Delete bucket ownership controls from S3 bucket
    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        None
    """
    client.delete_bucket_ownership_controls(Bucket=bucket_name)


# ========================================
# S3 Client Wrappers - Object Operations
# ========================================


@S3ErrorHandler.list_error_handler("get object metadata", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def head_s3_object(
    client: ClientType, bucket_name: str, object_key: str, version_id: Optional[str] = None, **kwargs
) -> Dict:
    """
    Retrieve metadata about an S3 object without downloading the object itself.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key of the S3 object.
        version_id (str, optional): The version ID of the object.
        **kwargs: Additional parameters to pass to head_object (e.g., ExpectedBucketOwner).

    Returns:
        Dict: Object metadata including ETag, ContentLength, LastModified, etc.
              Returns {} if object does not exist (404).
    """
    params = {"Bucket": bucket_name, "Key": object_key}
    if version_id:
        params["VersionId"] = version_id
    params.update(kwargs)
    return client.head_object(**params)


@AWSRetry.jittered_backoff()
def s3_head_objects(
    client: ClientType, parts: int, bucket: str, obj: str, versionId: Optional[str]
) -> Generator[Dict, None, None]:
    """
    Generator that yields HEAD object responses for each part of a multipart upload.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        parts (int): Number of parts in the multipart upload.
        bucket (str): The S3 bucket name.
        obj (str): The S3 object key.
        versionId (str): Optional version ID of the object.

    Yields:
        HEAD object response dict for each part.
    """
    args = {"Bucket": bucket, "Key": obj}
    if versionId:
        args["VersionId"] = versionId

    for part in range(1, parts + 1):
        args["PartNumber"] = part
        yield client.head_object(**args)


@S3ErrorHandler.list_error_handler("get object content", None)
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_object_content(
    client: ClientType, bucket_name: str, object_key: str, version_id: Optional[str] = None, **kwargs
) -> Optional[bytes]:
    """
    Download the content of an S3 object.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key of the S3 object.
        version_id (str, optional): The version ID of the object.
        **kwargs: Additional parameters to pass to get_object.

    Returns:
        bytes: The object content, or None if object does not exist (404).
    """
    params = {"Bucket": bucket_name, "Key": object_key}
    if version_id:
        params["VersionId"] = version_id
    params.update(kwargs)
    response = client.get_object(**params)
    return response["Body"].read()


@S3ErrorHandler.list_error_handler("get object tags", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_object_tagging(
    client: ClientType, bucket_name: str, object_key: str, version_id: Optional[str] = None, **kwargs
) -> Dict:
    """
    Retrieve tags for an S3 object.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key of the S3 object.
        version_id (str, optional): The version ID of the object.
        **kwargs: Additional parameters to pass to get_object_tagging (e.g., ExpectedBucketOwner).

    Returns:
        Dict: The current tags dictionary applied to the object.
              Returns {} if object has no tags or does not exist.
    """
    params = {"Bucket": bucket_name, "Key": object_key}
    if version_id:
        params["VersionId"] = version_id
    params.update(kwargs)
    current_tags = client.get_object_tagging(**params).get("TagSet", [])
    return boto3_tag_list_to_ansible_dict(current_tags)


@S3ErrorHandler.list_error_handler("get object ACL", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_object_acl(
    client: ClientType, bucket_name: str, object_key: str, version_id: Optional[str] = None, **kwargs
) -> Dict:
    """
    Retrieve ACL for an S3 object.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key of the S3 object.
        version_id (str, optional): The version ID of the object.
        **kwargs: Additional parameters to pass to get_object_acl.

    Returns:
        Dict: ACL information including Owner and Grants.
              Returns {} if object does not exist.
    """
    params = {"Bucket": bucket_name, "Key": object_key}
    if version_id:
        params["VersionId"] = version_id
    params.update(kwargs)
    return client.get_object_acl(**params)


@S3ErrorHandler.list_error_handler("get object legal hold status", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_object_legal_hold(
    client: ClientType, bucket_name: str, object_key: str, version_id: Optional[str] = None, **kwargs
) -> Dict:
    """
    Retrieve legal hold status for an S3 object.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key of the S3 object.
        version_id (str, optional): The version ID of the object.
        **kwargs: Additional parameters to pass to get_object_legal_hold.

    Returns:
        Dict: Legal hold configuration.
              Returns {} if not configured or object does not exist.
    """
    params = {"Bucket": bucket_name, "Key": object_key}
    if version_id:
        params["VersionId"] = version_id
    params.update(kwargs)
    result = client.get_object_legal_hold(**params)
    return result.get("LegalHold", {})


@S3ErrorHandler.list_error_handler("get object retention settings", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_object_retention(
    client: ClientType, bucket_name: str, object_key: str, version_id: Optional[str] = None, **kwargs
) -> Dict:
    """
    Retrieve retention settings for an S3 object.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key of the S3 object.
        version_id (str, optional): The version ID of the object.
        **kwargs: Additional parameters to pass to get_object_retention.

    Returns:
        Dict: Retention configuration.
              Returns {} if not configured or object does not exist.
    """
    params = {"Bucket": bucket_name, "Key": object_key}
    if version_id:
        params["VersionId"] = version_id
    params.update(kwargs)
    result = client.get_object_retention(**params)
    return result.get("Retention", {})


@S3ErrorHandler.list_error_handler("get object attributes", {})
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def get_s3_object_attributes(
    client: ClientType,
    bucket_name: str,
    object_key: str,
    object_attributes: List[str],
    version_id: Optional[str] = None,
    **kwargs,
) -> Dict:
    """
    Retrieve specific attributes for an S3 object.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key of the S3 object.
        object_attributes (list): List of object attributes to retrieve
                                  (e.g., ['ETag', 'Checksum', 'ObjectParts', 'StorageClass', 'ObjectSize']).
        version_id (str, optional): The version ID of the object.
        **kwargs: Additional parameters to pass to get_object_attributes.

    Returns:
        Dict: Requested object attributes.
              Returns {} if object does not exist.
    """
    params = {
        "Bucket": bucket_name,
        "Key": object_key,
        "ObjectAttributes": object_attributes,
    }
    if version_id:
        params["VersionId"] = version_id
    params.update(kwargs)
    return client.get_object_attributes(**params)


@AWSRetry.jittered_backoff()
def s3_object_exists(
    client: ClientType,
    bucket_name: str,
    object_key: str,
    version_id: Optional[str] = None,
) -> bool:
    """
    Check if an S3 object exists.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key of the S3 object.
        version_id (str, optional): The version ID of the object.

    Returns:
        bool: True if object exists, False otherwise.

    Raises:
        AnsibleS3PermissionsError: If access is denied (403).
        AnsibleS3Error: For other S3 errors.
    """
    result = head_s3_object(client, bucket_name, object_key, version_id)
    return bool(result)


@S3ErrorHandler.deletion_error_handler("delete object")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def delete_s3_object(client: ClientType, bucket_name: str, object_key: str, **kwargs) -> Dict:
    """
    Delete an S3 object.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key of the S3 object to delete.
        **kwargs: Additional parameters to pass to delete_object.

    Returns:
        Dict: Response from the delete operation.
    """
    params = {"Bucket": bucket_name, "Key": object_key}
    params.update(kwargs)
    return client.delete_object(**params)


@S3ErrorHandler.common_error_handler("update object tags")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_s3_object_tagging(
    client: ClientType,
    bucket_name: str,
    object_key: str,
    tags: Dict,
    **kwargs,
) -> Dict:
    """
    Set tags on an S3 object.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key of the S3 object.
        tags (dict): Dictionary of tags to apply (e.g., {'Environment': 'prod', 'Owner': 'team'}).
        **kwargs: Additional parameters to pass to put_object_tagging.

    Returns:
        Dict: Response from the put operation.
    """
    params = {
        "Bucket": bucket_name,
        "Key": object_key,
        "Tagging": {"TagSet": ansible_dict_to_boto3_tag_list(tags)},
    }
    params.update(kwargs)
    return client.put_object_tagging(**params)


@S3ErrorHandler.deletion_error_handler("delete object tags")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def delete_s3_object_tagging(client: ClientType, bucket_name: str, object_key: str, **kwargs) -> Dict:
    """
    Remove all tags from an S3 object.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key of the S3 object.
        **kwargs: Additional parameters to pass to delete_object_tagging.

    Returns:
        Dict: Response from the delete operation.
    """
    params = {"Bucket": bucket_name, "Key": object_key}
    params.update(kwargs)
    return client.delete_object_tagging(**params)


@S3ErrorHandler.common_error_handler("update object ACL")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_s3_object_acl(client: ClientType, bucket_name: str, object_key: str, acl: str, **kwargs) -> Dict:
    """
    Set ACL on an S3 object.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key of the S3 object.
        acl (str): The canned ACL to apply (e.g., 'private', 'public-read', 'authenticated-read').
        **kwargs: Additional parameters to pass to put_object_acl.

    Returns:
        Dict: Response from the put operation.
    """
    params = {"Bucket": bucket_name, "Key": object_key, "ACL": acl}
    params.update(kwargs)
    return client.put_object_acl(**params)


@S3ErrorHandler.common_error_handler("put object")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_s3_object(client: ClientType, **params) -> Dict:
    """
    Upload an object to S3 using put_object.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        **params: Parameters to pass to put_object (must include Bucket, Key, and typically Body).

    Returns:
        Dict: Response from the put operation.
    """
    return client.put_object(**params)


@S3ErrorHandler.common_error_handler("upload file")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def upload_s3_file(
    client: ClientType,
    filename: str,
    bucket_name: str,
    object_key: str,
    extra_args: Optional[Dict] = None,
) -> None:
    """
    Upload a file to S3 using upload_file (transfer manager).

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        filename (str): Path to the file to upload.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key for the S3 object.
        extra_args (dict, optional): Extra arguments for the upload (e.g., ContentType, Metadata).

    Returns:
        None
    """
    if extra_args is None:
        extra_args = {}
    client.upload_file(Filename=filename, Bucket=bucket_name, Key=object_key, ExtraArgs=extra_args)


@S3ErrorHandler.common_error_handler("download file")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def download_s3_file(
    client: ClientType,
    bucket_name: str,
    object_key: str,
    filename: str,
    extra_args: Optional[Dict] = None,
) -> None:
    """
    Download a file from S3 using download_file (transfer manager).

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key of the S3 object.
        filename (str): Path where the file should be saved.
        extra_args (dict, optional): Extra arguments for the download (e.g., VersionId).

    Returns:
        None
    """
    if extra_args is None:
        extra_args = {}
    client.download_file(Bucket=bucket_name, Key=object_key, Filename=filename, ExtraArgs=extra_args)


@S3ErrorHandler.common_error_handler("copy object")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def copy_s3_object(
    client: ClientType,
    copy_source: Dict,
    dest_bucket_name: str,
    dest_object_key: str,
    extra_args: Optional[Dict] = None,
) -> None:
    """
    Copy an object within S3 using managed copy (automatically handles multipart for large objects).

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        copy_source (dict): Source object specification with 'Bucket', 'Key', and optionally 'VersionId'.
        dest_bucket_name (str): The destination bucket name.
        dest_object_key (str): The destination object key.
        extra_args (dict, optional): Extra arguments for the copy operation (e.g., ACL, Metadata).

    Returns:
        None
    """
    if extra_args is None:
        extra_args = {}
    client.copy(copy_source, dest_bucket_name, dest_object_key, ExtraArgs=extra_args)


@AWSRetry.jittered_backoff()
def ensure_s3_object_tags(
    client: ClientType,
    bucket_name: str,
    object_key: str,
    desired_tags: Optional[Dict],
    purge_tags: bool = True,
    _max_attempts: int = 12,
    _sleep_time: float = 5.0,
) -> Tuple[Dict, bool]:
    """
    Ensure S3 object has desired tags, optionally purging unspecified tags.

    This function handles the complete tag management workflow:
    - Retrieves current tags
    - Compares with desired state
    - Updates or deletes tags as needed
    - Waits for tags to be applied

    Handles S3-compatible services that don't support tagging by returning empty tags
    when desired_tags is None.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key of the S3 object.
        desired_tags (dict or None): Dictionary of desired tags. If None, tags are not modified.
        purge_tags (bool): If True, remove tags not in desired_tags. If False, merge with existing tags.
        _max_attempts (int): Maximum retry attempts for eventual consistency (internal/testing parameter).
        _sleep_time (float): Sleep time between retries in seconds (internal/testing parameter).

    Returns:
        Tuple[Dict, bool]: (current_tags, changed) - The current tags and whether changes were made.

    Raises:
        AnsibleS3Error: If tag operations fail and desired_tags is not None.
        AnsibleS3SupportError: If tagging is not supported and desired_tags is not None.
    """
    import time

    # Get current tags, handling S3-compatible services that don't support tagging
    try:
        current_tags_dict = get_s3_object_tagging(client, bucket_name, object_key)
    except AnsibleS3SupportError:
        # If tags weren't requested, silently handle unsupported tagging
        if desired_tags is None:
            return {}, False
        # If tags were requested, re-raise to indicate tagging is not supported
        raise

    # If no tags were requested, return current tags without changes
    if desired_tags is None:
        return current_tags_dict, False

    # Compute final tags using existing merge_tags helper
    final_tags = merge_tags(current_tags_dict, desired_tags, purge_tags)

    # Nothing to change
    if current_tags_dict == final_tags:
        return current_tags_dict, False

    # Apply tag changes
    if final_tags:
        put_s3_object_tagging(client, bucket_name, object_key, final_tags)
    else:
        delete_s3_object_tagging(client, bucket_name, object_key)

    # Wait for tags to be applied (eventual consistency)
    for _attempt in range(_max_attempts):
        current_tags_dict = get_s3_object_tagging(client, bucket_name, object_key)
        if current_tags_dict == final_tags:
            return current_tags_dict, True
        time.sleep(_sleep_time)

    # Tags didn't apply in time, but return what we have
    return current_tags_dict, True


@S3ErrorHandler.common_error_handler("generate presigned URL")
@AWSRetry.jittered_backoff()
def generate_s3_presigned_url(
    client: ClientType,
    bucket_name: str,
    object_key: str,
    client_method: str,
    expiry: int = 600,
    **params,
) -> str:
    """
    Generate a presigned URL for S3 object operations.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        object_key (str): The key of the S3 object.
        client_method (str): The boto3 client method name ('get_object', 'put_object', etc.).
        expiry (int): URL expiration time in seconds (default: 600).
        **params: Additional parameters for the presigned URL.

    Returns:
        str: The presigned URL, or empty string on error.

    Raises:
        AnsibleS3Error: If URL generation fails.
    """
    url_params = {"Bucket": bucket_name, "Key": object_key}
    url_params.update(params)

    return client.generate_presigned_url(ClientMethod=client_method, Params=url_params, ExpiresIn=expiry)


@AWSRetry.jittered_backoff()
def _list_objects_v2(client: ClientType, **params) -> Dict:
    params = {k: v for k, v in params.items() if v is not None}
    # For practical purposes, the paginator ignores MaxKeys, if we've been passed MaxKeys we need to
    # explicitly call list_objects_v3 rather than re-use the paginator
    if params.get("MaxKeys", None) is not None:
        return client.list_objects_v2(**params)

    paginator = client.get_paginator("list_objects_v2")
    return paginator.paginate(**params).build_full_result()


@S3ErrorHandler.list_error_handler("list bucket objects", [])
def list_bucket_object_keys(
    client: ClientType,
    bucket: str,
    prefix: Optional[str] = None,
    max_keys: Optional[int] = None,
    start_after: Optional[str] = None,
) -> List[str]:
    """
    List object keys in an S3 bucket with optional filtering.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        bucket (str): The name of the S3 bucket.
        prefix (str, optional): Limits the response to keys that begin with the specified prefix.
        max_keys (int, optional): Maximum number of keys to return.
        start_after (str, optional): StartAfter key for pagination.

    Returns:
        List[str]: List of object keys. Returns [] if bucket has no objects or on 404.

    Raises:
        AnsibleS3PermissionsError: If access is denied (403).
        AnsibleS3Error: For other S3 errors.
    """
    response = _list_objects_v2(client, Bucket=bucket, Prefix=prefix, StartAfter=start_after, MaxKeys=max_keys)
    return [c["Key"] for c in response.get("Contents", [])]


# ==========================================
# Transformation and Helper Functions
# ==========================================


def calculate_checksum_with_file(
    client: ClientType, parts: int, bucket: str, obj: str, versionId: Optional[str], filename: str
) -> str:
    """
    Calculate MD5 checksum for a multipart upload using a local file.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        parts (int): Number of parts in the multipart upload.
        bucket (str): The S3 bucket name.
        obj (str): The S3 object key.
        versionId (str): Optional version ID of the object.
        filename (str): Path to the local file.

    Returns:
        The ETag for the multipart upload in the format "{md5hash}-{parts}".
    """
    digests = []
    with open(filename, "rb") as f:
        for head in s3_head_objects(client, parts, bucket, obj, versionId):
            digests.append(md5(f.read(int(head["ContentLength"]))).digest())

    digest_squared = b"".join(digests)
    return f'"{md5(digest_squared).hexdigest()}-{len(digests)}"'


def calculate_checksum_with_content(
    client: ClientType, parts: int, bucket: str, obj: str, versionId: Optional[str], content: bytes
) -> str:
    """
    Calculate MD5 checksum for a multipart upload using in-memory content.

    Parameters:
        client (boto3.client): The Boto3 S3 client object.
        parts (int): Number of parts in the multipart upload.
        bucket (str): The S3 bucket name.
        obj (str): The S3 object key.
        versionId (str): Optional version ID of the object.
        content (bytes): The content bytes.

    Returns:
        The ETag for the multipart upload in the format "{md5hash}-{parts}".
    """
    digests = []
    offset = 0
    for head in s3_head_objects(client, parts, bucket, obj, versionId):
        length = int(head["ContentLength"])
        digests.append(md5(content[offset:offset + length]).digest())  # fmt: skip
        offset += length

    digest_squared = b"".join(digests)
    return f'"{md5(digest_squared).hexdigest()}-{len(digests)}"'


def calculate_etag(
    module: AnsibleAWSModule,
    filename: str,
    etag: Optional[str],
    client: ClientType,
    bucket: str,
    obj: str,
    version: Optional[str] = None,
) -> Optional[str]:
    """
    Calculate the ETag for a local file, handling both single-part and multipart uploads.

    Parameters:
        module (AnsibleAWSModule): The Ansible module object.
        filename (str): Path to the local file.
        etag (str): The ETag from S3 to determine if multipart.
        client (boto3.client): The Boto3 S3 client object.
        bucket (str): The S3 bucket name.
        obj (str): The S3 object key.
        version (str): Optional version ID of the object.

    Returns:
        The calculated ETag string, or None if MD5 is not available.
    """
    if not HAS_MD5:
        return None

    if etag is not None and "-" in etag:
        # Multi-part ETag; a hash of the hashes of each part.
        parts = int(etag[1:-1].split("-")[1])
        try:
            return calculate_checksum_with_file(client, parts, bucket, obj, version, filename)
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
            raise AnsibleS3Error(message="Failed to calculate object ETag", exception=e) from e
    else:  # Compute the MD5 sum normally
        return f'"{module.md5(filename)}"'


def calculate_etag_content(
    module: AnsibleAWSModule,
    content: bytes,
    etag: Optional[str],
    client: ClientType,
    bucket: str,
    obj: str,
    version: Optional[str] = None,
) -> Optional[str]:
    """
    Calculate the ETag for in-memory content, handling both single-part and multipart uploads.

    Parameters:
        module (AnsibleAWSModule): The Ansible module object.
        content (bytes): The content bytes.
        etag (str): The ETag from S3 to determine if multipart.
        client (boto3.client): The Boto3 S3 client object.
        bucket (str): The S3 bucket name.
        obj (str): The S3 object key.
        version (str): Optional version ID of the object.

    Returns:
        The calculated ETag string, or None if MD5 is not available.
    """
    if not HAS_MD5:
        return None

    if etag is not None and "-" in etag:
        # Multi-part ETag; a hash of the hashes of each part.
        parts = int(etag[1:-1].split("-")[1])
        try:
            return calculate_checksum_with_content(client, parts, bucket, obj, version, content)
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
            raise AnsibleS3Error(message="Failed to calculate object ETag", exception=e) from e
    else:  # Compute the MD5 sum normally
        return f'"{md5(content).hexdigest()}"'


def validate_bucket_name(name: str) -> Optional[str]:
    """
    Validate S3 bucket name against AWS naming rules.

    Parameters:
        name (str): The bucket name to validate.

    Returns:
        Error message string if validation fails, or None if valid.
    """
    # See: https://docs.aws.amazon.com/AmazonS3/latest/userguide/bucketnamingrules.html
    if len(name) < 3:
        return "the length of an S3 bucket must be at least 3 characters"
    if len(name) > 63:
        return "the length of an S3 bucket cannot exceed 63 characters"

    legal_characters = string.ascii_lowercase + ".-" + string.digits
    illegal_characters = [c for c in name if c not in legal_characters]
    if illegal_characters:
        return "invalid character(s) found in the bucket name"
    if name[-1] not in string.ascii_lowercase + string.digits:
        return "bucket names must begin and end with a letter or number"
    return None


def validate_account_regional_bucket_name(name: str, region: Optional[str] = None) -> Optional[str]:
    """
    Validate S3 bucket name for account-regional namespace.

    Account-regional bucket names must follow the format:
    prefix-{accountId}-{region}-an

    Parameters:
        name (str): The bucket name to validate.
        region (str): The AWS region (optional, used for validation if provided).

    Returns:
        Error message string if validation fails, or None if valid.
    """
    import re

    if len(name) < 3:
        return "account-regional bucket name must be at least 3 characters"
    if len(name) > 63:
        return "account-regional bucket name cannot exceed 63 characters"

    if not name.endswith("-an"):
        return "account-regional bucket name must end with '-an'"

    pattern = r"^(.+)-(\d{12})-([a-z]{2}-[a-z]+-\d)-an$"
    match = re.match(pattern, name)

    if not match:
        return "account-regional bucket name must follow format: prefix-{12-digit-account-id}-{region}-an"

    prefix, account_id, bucket_region = match.groups()

    if not prefix:
        return "account-regional bucket name must have a prefix before the account ID"

    if len(account_id) != 12:
        return "account ID in bucket name must be exactly 12 digits"

    if region and bucket_region != region:
        return f"bucket name region '{bucket_region}' does not match the bucket's configured region '{region}'"

    return None


# Spot special case of fakes3.
def is_fakes3(url: Optional[str]) -> bool:
    """Return True if endpoint_url has scheme fakes3://"""
    result = False
    if url is not None:
        result = urlparse(url).scheme in ("fakes3", "fakes3s")
    return result


def parse_fakes3_endpoint(url: str) -> Dict:
    """
    Parse a FakeS3 endpoint URL and extract connection parameters.

    Parameters:
        url (str): The FakeS3 endpoint URL (fakes3:// or fakes3s://).

    Returns:
        Dict with 'endpoint' URL and 'use_ssl' boolean.
    """
    fakes3 = urlparse(url)
    protocol = "http"
    port = fakes3.port or 80
    if fakes3.scheme == "fakes3s":
        protocol = "https"
        port = fakes3.port or 443
    endpoint_url = f"{protocol}://{fakes3.hostname}:{to_text(port)}"
    use_ssl = bool(fakes3.scheme == "fakes3s")
    return {"endpoint": endpoint_url, "use_ssl": use_ssl}


def parse_ceph_endpoint(url: str) -> Dict:
    """
    Parse a Ceph endpoint URL and extract connection parameters.

    Parameters:
        url (str): The Ceph endpoint URL.

    Returns:
        Dict with 'endpoint' URL and 'use_ssl' boolean.
    """
    ceph = urlparse(url)
    use_ssl = bool(ceph.scheme == "https")
    return {"endpoint": url, "use_ssl": use_ssl}


def parse_s3_endpoint(options: Dict) -> Tuple[bool, Dict]:
    """
    Parse S3 endpoint options and determine connection parameters.

    Parameters:
        options (dict): Dictionary containing endpoint_url, ceph flag, etc.

    Returns:
        Tuple of (is_aws_s3, connection_params_dict).
    """
    endpoint_url = options.get("endpoint_url")
    if options.get("ceph"):
        return False, parse_ceph_endpoint(endpoint_url)
    if is_fakes3(endpoint_url):
        return False, parse_fakes3_endpoint(endpoint_url)
    return True, {"endpoint": endpoint_url}


def s3_extra_params(options: Dict, sigv4: bool = False) -> Dict:
    """
    Build extra parameters for S3 client connection.

    Parameters:
        options (dict): Dictionary containing connection options.
        sigv4 (bool): Whether to use signature version 4.

    Returns:
        Dict of extra connection parameters for boto3 client.
    """
    aws, extra_params = parse_s3_endpoint(options)
    if not aws:
        return extra_params
    dualstack = options.get("dualstack")
    if not dualstack and not sigv4:
        return extra_params
    config: dict = {}
    if dualstack:
        config["use_dualstack_endpoint"] = True
    if sigv4:
        config["signature_version"] = "s3v4"
    extra_params["config"] = config
    return extra_params


def get_s3_bucket_location(module: AnsibleAWSModule) -> str:
    """
    Determine the S3 bucket location/region from module parameters.

    Parameters:
        module (AnsibleAWSModule): The Ansible module object.

    Returns:
        The region string for the S3 bucket location.
    """
    if module.params.get("ceph") is True:
        return module.params.get("region")
    return module.region or "us-east-1"
